from pathlib import Path
from typing import Any

import pytest

from litestar import get
from litestar.contrib.htmx._utils import HTMXHeaders
from litestar.contrib.htmx.request import HTMXRequest
from litestar.contrib.htmx.response import (
    ClientRedirect,
    ClientRefresh,
    HTMXTemplate,
    HXLocation,
    HXStopPolling,
    PushUrl,
    ReplaceUrl,
    Reswap,
    Retarget,
    TriggerEvent,
)
from litestar.contrib.jinja import JinjaTemplateEngine
from litestar.contrib.mako import MakoTemplateEngine
from litestar.status_codes import HTTP_200_OK, HTTP_500_INTERNAL_SERVER_ERROR
from litestar.template.config import TemplateConfig
from litestar.testing import create_test_client


async def test_hx_stop_polling_response() -> None:
    @get("/")
    def handler() -> HXStopPolling:
        return HXStopPolling()

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get("/")
        assert response.status_code == 286


async def test_client_redirect_response() -> None:
    @get("/")
    def handler() -> ClientRedirect:
        return ClientRedirect(redirect_to="https://example.com")

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get("/")
        assert response.status_code == HTTP_200_OK
        assert response.headers.get(HTMXHeaders.REDIRECT) == "https://example.com"
        assert response.headers.get("location") is None


async def test_client_refresh_response() -> None:
    @get("/")
    def handler() -> ClientRefresh:
        return ClientRefresh()

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get("/")
        assert response.status_code == HTTP_200_OK
        assert response.headers[HTMXHeaders.REFRESH] == "true"


async def test_push_url_false_response() -> None:
    @get("/")
    def handler() -> PushUrl:
        return PushUrl(content="Success!", push_url=False)

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get("/")
        assert response.status_code == HTTP_200_OK
        assert response.headers[HTMXHeaders.PUSH_URL] == "false"


async def test_push_url_response() -> None:
    @get("/")
    def handler() -> PushUrl:
        return PushUrl(content="Success!", push_url="/index.html")

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get("/")
        assert response.status_code == HTTP_200_OK
        assert response.text == "Success!"
        assert response.headers[HTMXHeaders.PUSH_URL] == "/index.html"


async def test_replace_url_false_response() -> None:
    @get("/")
    def handler() -> ReplaceUrl:
        return ReplaceUrl(content="Success!", replace_url=False)

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get("/")
        assert response.status_code == HTTP_200_OK
        assert response.headers[HTMXHeaders.REPLACE_URL] == "false"


async def test_replace_url_response() -> None:
    @get("/")
    def handler() -> ReplaceUrl:
        return ReplaceUrl(content="Success!", replace_url="/index.html")

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get("/")
        assert response.status_code == HTTP_200_OK
        assert response.text == "Success!"
        assert response.headers[HTMXHeaders.REPLACE_URL] == "/index.html"


async def test_reswap_response() -> None:
    @get("/")
    def handler() -> Reswap:
        return Reswap(content="Success!", method="beforebegin")

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get("/")
        assert response.status_code == HTTP_200_OK
        assert response.text == "Success!"
        assert response.headers[HTMXHeaders.RE_SWAP] == "beforebegin"


async def test_retarget_response() -> None:
    @get("/")
    def handler() -> Retarget:
        return Retarget(content="Success!", target="#element")

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get("/")
        assert response.status_code == HTTP_200_OK
        assert response.text == "Success!"
        assert response.headers[HTMXHeaders.RE_TARGET] == "#element"


async def test_trigger_event_response_success() -> None:
    @get("/")
    def handler() -> TriggerEvent:
        return TriggerEvent(
            content="Success!", name="alert", after="receive", params={"warning": "Confirm your choice!"}
        )

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get("/")
        assert response.status_code == HTTP_200_OK
        assert response.text == "Success!"
        assert response.headers[HTMXHeaders.TRIGGER_EVENT] == '{"alert":{"warning":"Confirm your choice!"}}'


async def test_trigger_event_response_no_params() -> None:
    @get("/")
    def handler() -> TriggerEvent:
        return TriggerEvent(content="Success!", name="alert", after="receive")

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get("/")

        assert response.status_code == HTTP_200_OK
        assert response.text == "Success!"
        assert response.headers[HTMXHeaders.TRIGGER_EVENT] == '{"alert":{}}'


async def test_trigger_event_response_after_settle() -> None:
    @get("/")
    def handler() -> TriggerEvent:
        return TriggerEvent(
            content="Success!", name="alert", after="settle", params={"warning": "Confirm your choice!"}
        )

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get("/")
        assert response.status_code == HTTP_200_OK
        assert response.text == "Success!"
        assert response.headers[HTMXHeaders.TRIGGER_AFTER_SETTLE] == '{"alert":{"warning":"Confirm your choice!"}}'


async def test_trigger_event_response_after_swap() -> None:
    @get("/")
    def handler() -> TriggerEvent:
        return TriggerEvent(content="Success!", name="alert", after="swap", params={"warning": "Confirm your choice!"})

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get("/")
        assert response.status_code == HTTP_200_OK
        assert response.text == "Success!"
        assert response.headers[HTMXHeaders.TRIGGER_AFTER_SWAP] == '{"alert":{"warning":"Confirm your choice!"}}'


async def test_trigger_event_response_invalid_after() -> None:
    @get("/")
    def handler() -> TriggerEvent:
        return TriggerEvent(
            content="Success!",
            name="alert",
            after="invalid",  # type: ignore[arg-type]
            params={"warning": "Confirm your choice!"},
        )

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get("/")
        assert response.status_code == HTTP_500_INTERNAL_SERVER_ERROR


async def test_hx_location_response_success() -> None:
    @get("/")
    def handler() -> HXLocation:
        return HXLocation(redirect_to="/contact-us")

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get("/")
        spec = response.headers[HTMXHeaders.LOCATION]
        assert response.status_code == HTTP_200_OK
        assert "Location" not in response.headers
        assert spec == '{"path":"/contact-us"}'


async def test_hx_location_response_with_all_parameters() -> None:
    @get("/")
    def handler() -> HXLocation:
        return HXLocation(
            redirect_to="/contact-us",
            source="#button",
            event="click",
            target="#content",
            swap="innerHTML",
            hx_headers={"attribute": "value"},
            values={"action": "true"},
        )

    with create_test_client(route_handlers=[handler], request_class=HTMXRequest) as client:
        response = client.get("/")
        spec = response.headers[HTMXHeaders.LOCATION]
        assert response.status_code == HTTP_200_OK
        assert "Location" not in response.headers
        assert spec == (
            '{"path":"/contact-us","source":"#button","event":"click","target":"#content","swap":"innerHTML",'
            '"values":{"action":"true"},"hx_headers":{"attribute":"value"}}'
        )


@pytest.mark.parametrize(
    "engine, template, expected",
    (
        (
            JinjaTemplateEngine,
            "path: {{ request.scope['path'] }} custom_key: {{ custom_key }}",
            "path: / custom_key: custom_value",
        ),
        (
            MakoTemplateEngine,
            "path: ${request.scope['path']} custom_key: ${custom_key}",
            "path: / custom_key: custom_value",
        ),
    ),
)
def test_HTMXTemplate_response_success(engine: Any, template: str, expected: str, tmp_path: Path) -> None:
    Path(tmp_path / "abc.html").write_text(template)

    @get(path="/")
    def handler() -> HTMXTemplate:
        return HTMXTemplate(
            template_name="abc.html",
            context={"request": {"scope": {"path": "nope"}}, "custom_key": "custom_value"},
            push_url="/about",
            re_swap="beforebegin",
            re_target="#new-target-id",
            trigger_event="showMessage",
            params={"alert": "Confirm your Choice."},
            after="receive",
        )

    with create_test_client(
        route_handlers=[handler],
        template_config=TemplateConfig(
            directory=tmp_path,
            engine=engine,
        ),
    ) as client:
        response = client.get("/")
        assert response.text == expected
        assert response.headers.get(HTMXHeaders.PUSH_URL) == "/about"
        assert response.headers.get(HTMXHeaders.RE_SWAP) == "beforebegin"
        assert response.headers.get(HTMXHeaders.RE_TARGET) == "#new-target-id"
        assert response.headers.get(HTMXHeaders.TRIGGER_EVENT) == '{"showMessage":{"alert":"Confirm your Choice."}}'


@pytest.mark.parametrize(
    "engine, template, expected",
    (
        (JinjaTemplateEngine, "path: {{ request.scope['path'] }}", "path: /"),
        (MakoTemplateEngine, "path: ${request.scope['path']}", "path: /"),
    ),
)
def test_HTMXTemplate_response_no_params(engine: Any, template: str, expected: str, tmp_path: Path) -> None:
    Path(tmp_path / "abc.html").write_text(template)

    @get(path="/")
    def handler() -> HTMXTemplate:
        return HTMXTemplate(
            template_name="abc.html",
            context={"request": {"scope": {"path": "nope"}}},
        )

    with create_test_client(
        route_handlers=[handler],
        template_config=TemplateConfig(
            directory=tmp_path,
            engine=engine,
        ),
    ) as client:
        response = client.get("/")
        assert response.text == expected
        assert response.headers.get(HTMXHeaders.PUSH_URL) is None
        assert response.headers.get(HTMXHeaders.RE_SWAP) is None
        assert response.headers.get(HTMXHeaders.RE_TARGET) is None
        assert response.headers.get(HTMXHeaders.TRIGGER_EVENT) is None


@pytest.mark.parametrize(
    "engine, template, expected",
    (
        (JinjaTemplateEngine, "path: {{ request.scope['path'] }}", "path: /"),
        (MakoTemplateEngine, "path: ${request.scope['path']}", "path: /"),
    ),
)
def test_HTMXTemplate_response_push_url_set_to_false(engine: Any, template: str, expected: str, tmp_path: Path) -> None:
    Path(tmp_path / "abc.html").write_text(template)

    @get(path="/")
    def handler() -> HTMXTemplate:
        return HTMXTemplate(
            template_name="abc.html",
            context={"request": {"scope": {"path": "nope"}}},
            push_url=False,
        )

    with create_test_client(
        route_handlers=[handler],
        template_config=TemplateConfig(
            directory=tmp_path,
            engine=engine,
        ),
    ) as client:
        response = client.get("/")
        assert response.text == expected
        assert response.headers.get(HTMXHeaders.PUSH_URL) == "false"
        assert response.headers.get(HTMXHeaders.RE_SWAP) is None
        assert response.headers.get(HTMXHeaders.RE_TARGET) is None
        assert response.headers.get(HTMXHeaders.TRIGGER_EVENT) is None


@pytest.mark.parametrize(
    "engine, template, expected",
    (
        (JinjaTemplateEngine, "path: {{ request.scope['path'] }}", "path: /"),
        (MakoTemplateEngine, "path: ${request.scope['path']}", "path: /"),
    ),
)
def test_htmx_template_response_bad_trigger_params(engine: Any, template: str, expected: str, tmp_path: Path) -> None:
    Path(tmp_path / "abc.html").write_text(template)

    @get(path="/")
    def handler() -> HTMXTemplate:
        return HTMXTemplate(
            template_name="abc.html",
            context={"request": {"scope": {"path": "nope"}}},
            trigger_event="showMessage",
            params={"alert": "Confirm your Choice."},
            after="begin",  # type: ignore[arg-type]
        )

    with create_test_client(
        route_handlers=[handler],
        template_config=TemplateConfig(
            directory=tmp_path,
            engine=engine,
        ),
    ) as client:
        response = client.get("/")
        assert response.status_code == HTTP_500_INTERNAL_SERVER_ERROR
        assert response.headers.get(HTMXHeaders.PUSH_URL) is None
        assert response.headers.get(HTMXHeaders.RE_SWAP) is None
        assert response.headers.get(HTMXHeaders.RE_TARGET) is None
        assert response.headers.get(HTMXHeaders.TRIGGER_EVENT) is None



from docs.examples.plugins.di_plugin import app

from litestar.testing import TestClient


def test_di_plugin_example() -> None:
    with TestClient(app) as client:
        res = client.get("/?param=hello")
        assert res.status_code == 200
        assert res.text == "hello"



from .asgi_types import (
    ASGIApp,
    ASGIVersion,
    BaseScope,
    HTTPDisconnectEvent,
    HTTPReceiveMessage,
    HTTPRequestEvent,
    HTTPResponseBodyEvent,
    HTTPResponseStartEvent,
    HTTPScope,
    HTTPSendMessage,
    HTTPServerPushEvent,
    LifeSpanReceive,
    LifeSpanReceiveMessage,
    LifeSpanScope,
    LifeSpanSend,
    LifeSpanSendMessage,
    LifeSpanShutdownCompleteEvent,
    LifeSpanShutdownEvent,
    LifeSpanShutdownFailedEvent,
    LifeSpanStartupCompleteEvent,
    LifeSpanStartupEvent,
    LifeSpanStartupFailedEvent,
    Message,
    Method,
    Receive,
    ReceiveMessage,
    Scope,
    ScopeSession,
    Send,
    WebSocketAcceptEvent,
    WebSocketCloseEvent,
    WebSocketConnectEvent,
    WebSocketDisconnectEvent,
    WebSocketReceiveEvent,
    WebSocketReceiveMessage,
    WebSocketResponseBodyEvent,
    WebSocketResponseStartEvent,
    WebSocketScope,
    WebSocketSendEvent,
    WebSocketSendMessage,
)
from .builtin_types import TypedDictClass
from .callable_types import (
    AfterExceptionHookHandler,
    AfterRequestHookHandler,
    AfterResponseHookHandler,
    AnyCallable,
    AnyGenerator,
    AsyncAnyCallable,
    BeforeMessageSendHookHandler,
    BeforeRequestHookHandler,
    CacheKeyBuilder,
    ExceptionHandler,
    GetLogger,
    Guard,
    LifespanHook,
    OnAppInitHandler,
    OperationIDCreator,
    Serializer,
)
from .composite_types import (
    Dependencies,
    ExceptionHandlersMap,
    Middleware,
    ParametersMap,
    PathType,
    ResponseCookies,
    ResponseHeaders,
    Scopes,
    TypeDecodersSequence,
    TypeEncodersMap,
)
from .empty import Empty, EmptyType
from .file_types import FileInfo, FileSystemProtocol
from .helper_types import AnyIOBackend, MaybePartial, OptionalSequence, SSEData, StreamType, SyncOrAsyncUnion
from .internal_types import ControllerRouterHandler, ReservedKwargs, RouteHandlerMapItem, RouteHandlerType
from .protocols import DataclassProtocol, Logger
from .serialization import DataContainerType, LitestarEncodableType

__all__ = (
    "ASGIApp",
    "ASGIVersion",
    "AfterExceptionHookHandler",
    "AfterRequestHookHandler",
    "AfterResponseHookHandler",
    "AnyCallable",
    "AnyGenerator",
    "AnyIOBackend",
    "AsyncAnyCallable",
    "BaseScope",
    "BeforeMessageSendHookHandler",
    "BeforeRequestHookHandler",
    "CacheKeyBuilder",
    "ControllerRouterHandler",
    "DataContainerType",
    "DataclassProtocol",
    "Dependencies",
    "Empty",
    "EmptyType",
    "ExceptionHandler",
    "ExceptionHandlersMap",
    "FileInfo",
    "FileSystemProtocol",
    "GetLogger",
    "Guard",
    "HTTPDisconnectEvent",
    "HTTPReceiveMessage",
    "HTTPRequestEvent",
    "HTTPResponseBodyEvent",
    "HTTPResponseStartEvent",
    "HTTPScope",
    "HTTPSendMessage",
    "HTTPServerPushEvent",
    "LifeSpanReceive",
    "LifeSpanReceiveMessage",
    "LifeSpanScope",
    "LifeSpanSend",
    "LifeSpanSendMessage",
    "LifeSpanShutdownCompleteEvent",
    "LifeSpanShutdownEvent",
    "LifeSpanShutdownFailedEvent",
    "LifeSpanStartupCompleteEvent",
    "LifeSpanStartupEvent",
    "LifeSpanStartupFailedEvent",
    "LifespanHook",
    "LitestarEncodableType",
    "Logger",
    "MaybePartial",
    "Message",
    "Method",
    "Middleware",
    "OnAppInitHandler",
    "OperationIDCreator",
    "OptionalSequence",
    "ParametersMap",
    "PathType",
    "Receive",
    "ReceiveMessage",
    "ReservedKwargs",
    "ResponseCookies",
    "ResponseHeaders",
    "RouteHandlerMapItem",
    "RouteHandlerType",
    "Scope",
    "ScopeSession",
    "Scopes",
    "Send",
    "Serializer",
    "StreamType",
    "SSEData",
    "SyncOrAsyncUnion",
    "TypeDecodersSequence",
    "TypeEncodersMap",
    "TypedDictClass",
    "WebSocketAcceptEvent",
    "WebSocketCloseEvent",
    "WebSocketConnectEvent",
    "WebSocketDisconnectEvent",
    "WebSocketReceiveEvent",
    "WebSocketReceiveMessage",
    "WebSocketReceiveMessage",
    "WebSocketResponseBodyEvent",
    "WebSocketResponseStartEvent",
    "WebSocketScope",
    "WebSocketSendEvent",
    "WebSocketSendMessage",
    "WebSocketSendMessage",
)



from typing import Any

import pytest
from docs.examples.todo_app import full_app
from docs.examples.todo_app import update as update_app
from docs.examples.todo_app.create import dataclass as dataclass_create_app
from docs.examples.todo_app.create import dict as dict_create_app
from docs.examples.todo_app.get_list import dataclass as get_dataclass_app
from docs.examples.todo_app.get_list import dict as get_dict_app
from docs.examples.todo_app.get_list import (
    query_param,
    query_param_default,
    query_param_validate,
    query_param_validate_manually,
)
from msgspec import to_builtins

from litestar.testing import TestClient


@pytest.mark.parametrize("module", [update_app, full_app])
def test_update(module: Any) -> None:
    with TestClient(module.app) as client:
        res = client.put("/Profit", json={"title": "Profit", "done": True})

        assert res.status_code == 200
        assert module.TODO_LIST[2].done
        assert res.json() == to_builtins(module.TODO_LIST)


@pytest.mark.parametrize("module", [get_dataclass_app, get_dict_app, query_param_default])
def test_get_list_dataclass(module) -> None:
    with TestClient(module.app) as client:
        res = client.get("/")

        assert res.status_code == 200
        assert res.json() == to_builtins(module.TODO_LIST)


@pytest.mark.parametrize("module", [query_param, query_param_validate_manually, query_param_validate])
def test_get_list_query_param(module) -> None:
    with TestClient(module.app) as client:
        res = client.get("/?done=1")

        assert res.status_code == 200
        assert res.json() == to_builtins([i for i in module.TODO_LIST if i.done])


@pytest.mark.parametrize("module", [query_param_validate_manually, query_param_validate, full_app])
def test_get_list_query_param_invalid(module) -> None:
    with TestClient(module.app) as client:
        res = client.get("/?done=john")

        assert res.status_code == 400


def test_dict_create() -> None:
    with TestClient(dict_create_app.app) as client:
        res = client.post("/", json={"title": "foo", "done": True})

        assert res.status_code == 201
        assert res.json() == [{"title": "foo", "done": True}]
        assert [{"title": "foo", "done": True}] == dict_create_app.TODO_LIST


@pytest.mark.parametrize("module", [dataclass_create_app, full_app])
def test_dataclass_create(module: Any) -> None:
    with TestClient(module.app) as client:
        res = client.post("/", json={"title": "foo", "done": True})

        assert res.status_code == 201
        assert res.json() == to_builtins(module.TODO_LIST)
        assert len(module.TODO_LIST)
        assert module.TODO_LIST[-1] == module.TodoItem(title="foo", done=True)



import pydantic

from litestar import get
from litestar.testing import create_test_client


def test_app_with_v1_and_v2_models() -> None:
    class ModelV1(pydantic.v1.BaseModel):  # pyright: ignore
        foo: str

    class ModelV2(pydantic.BaseModel):
        foo: str

    @get("/v1")
    def handler_v1() -> ModelV1:
        return ModelV1(foo="bar")

    @get("/v2")
    def handler_v2() -> ModelV2:
        return ModelV2(foo="bar")

    with create_test_client([handler_v1, handler_v2]) as client:
        assert client.get("/v1").json() == {"foo": "bar"}
        assert client.get("/v2").json() == {"foo": "bar"}



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Mapping

from litestar.connection import WebSocket
from litestar.exceptions import ImproperlyConfiguredException
from litestar.handlers import BaseRouteHandler
from litestar.types.builtin_types import NoneType
from litestar.utils.predicates import is_async_callable

if TYPE_CHECKING:
    from litestar.types import Dependencies, ExceptionHandler, Guard, Middleware


class WebsocketRouteHandler(BaseRouteHandler):
    """Websocket route handler decorator.

    Use this decorator to decorate websocket handler functions.
    """

    __slots__ = ("websocket_class",)

    def __init__(
        self,
        path: str | list[str] | None = None,
        *,
        dependencies: Dependencies | None = None,
        exception_handlers: dict[int | type[Exception], ExceptionHandler] | None = None,
        guards: list[Guard] | None = None,
        middleware: list[Middleware] | None = None,
        name: str | None = None,
        opt: dict[str, Any] | None = None,
        signature_namespace: Mapping[str, Any] | None = None,
        websocket_class: type[WebSocket] | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize ``WebsocketRouteHandler``

        Args:
            path: A path fragment for the route handler function or a sequence of path fragments. If not given defaults
                to ``/``
            dependencies: A string keyed mapping of dependency :class:`Provider <.di.Provide>` instances.
            exception_handlers: A mapping of status codes and/or exception types to handler functions.
            guards: A sequence of :class:`Guard <.types.Guard>` callables.
            middleware: A sequence of :class:`Middleware <.types.Middleware>`.
            name: A string identifying the route handler.
            opt: A string keyed mapping of arbitrary values that can be accessed in :class:`Guards <.types.Guard>` or
                wherever you have access to :class:`Request <.connection.Request>` or
                :class:`ASGI Scope <.types.Scope>`.
            signature_namespace: A mapping of names to types for use in forward reference resolution during signature modelling.
            type_encoders: A mapping of types to callables that transform them into types supported for serialization.
            **kwargs: Any additional kwarg - will be set in the opt dictionary.
            websocket_class: A custom subclass of :class:`WebSocket <.connection.WebSocket>` to be used as route handler's
                default websocket class.
        """
        self.websocket_class = websocket_class

        super().__init__(
            path=path,
            dependencies=dependencies,
            exception_handlers=exception_handlers,
            guards=guards,
            middleware=middleware,
            name=name,
            opt=opt,
            signature_namespace=signature_namespace,
            **kwargs,
        )

    def resolve_websocket_class(self) -> type[WebSocket]:
        """Return the closest custom WebSocket class in the owner graph or the default Websocket class.

        This method is memoized so the computation occurs only once.

        Returns:
            The default :class:`WebSocket <.connection.WebSocket>` class for the route handler.
        """
        return next(
            (layer.websocket_class for layer in reversed(self.ownership_layers) if layer.websocket_class is not None),
            WebSocket,
        )

    def _validate_handler_function(self) -> None:
        """Validate the route handler function once it's set by inspecting its return annotations."""
        super()._validate_handler_function()

        if not self.parsed_fn_signature.return_type.is_subclass_of(NoneType):
            raise ImproperlyConfiguredException("Websocket handler functions should return 'None'")

        if "socket" not in self.parsed_fn_signature.parameters:
            raise ImproperlyConfiguredException("Websocket handlers must set a 'socket' kwarg")

        for param in ("request", "body", "data"):
            if param in self.parsed_fn_signature.parameters:
                raise ImproperlyConfiguredException(f"The {param} kwarg is not supported with websocket handlers")

        if not is_async_callable(self.fn):
            raise ImproperlyConfiguredException("Functions decorated with 'websocket' must be async functions")


websocket = WebsocketRouteHandler



"""A large part of the tests in this file were adapted from:

https://github.com/encode/starlette/blob/master/tests/test_responses.py And are meant to ensure our compatibility with
their API.
"""

from typing import TYPE_CHECKING, Optional

import pytest

from litestar import get
from litestar.exceptions import ImproperlyConfiguredException
from litestar.response.base import ASGIResponse
from litestar.response.redirect import ASGIRedirectResponse, Redirect
from litestar.status_codes import HTTP_200_OK
from litestar.testing import TestClient, create_test_client

if TYPE_CHECKING:
    from litestar.types import Receive, Scope, Send


def test_redirect_response() -> None:
    async def app(scope: "Scope", receive: "Receive", send: "Send") -> None:
        if scope["path"] == "/":
            response = ASGIResponse(body=b"hello, world", media_type="text/plain")
        else:
            response = ASGIRedirectResponse(path="/")
        await response(scope, receive, send)

    client = TestClient(app)
    response = client.get("/redirect")
    assert response.text == "hello, world"
    assert response.url == "http://testserver.local/"


def test_quoting_redirect_response() -> None:
    async def app(scope: "Scope", receive: "Receive", send: "Send") -> None:
        if scope["path"] == "/test/":
            response = ASGIResponse(body=b"hello, world", media_type="text/plain")
        else:
            response = ASGIRedirectResponse(path="/test/")
        await response(scope, receive, send)

    client = TestClient(app)
    response = client.get("/redirect", follow_redirects=True)
    assert response.text == "hello, world"
    assert str(response.url) == "http://testserver.local/test/"


def test_redirect_response_content_length_header() -> None:
    async def app(scope: "Scope", receive: "Receive", send: "Send") -> None:
        if scope["path"] == "/":
            response = ASGIResponse(body=b"hello", media_type="text/plain")
        else:
            response = ASGIRedirectResponse(path="/")
        await response(scope, receive, send)

    client: TestClient = TestClient(app)
    response = client.request("GET", "/redirect", follow_redirects=False)
    assert str(response.url) == "http://testserver.local/redirect"
    assert "content-length" in response.headers


def test_redirect_response_status_validation() -> None:
    with pytest.raises(ImproperlyConfiguredException):
        ASGIRedirectResponse(path="/", status_code=HTTP_200_OK)  # type:ignore[arg-type]


def test_redirect_response_html_media_type() -> None:
    async def app(scope: "Scope", receive: "Receive", send: "Send") -> None:
        if scope["path"] == "/":
            response = ASGIResponse(body=b"hello")
        else:
            response = ASGIRedirectResponse(path="/", media_type="text/html")
        await response(scope, receive, send)

    client: TestClient = TestClient(app)
    response = client.request("GET", "/redirect", follow_redirects=False)
    assert str(response.url) == "http://testserver.local/redirect"
    assert "text/html" in str(response.headers["Content-Type"])


def test_redirect_response_media_type_validation() -> None:
    with pytest.raises(ImproperlyConfiguredException):
        ASGIRedirectResponse(path="/", media_type="application/mspgpack")


@pytest.mark.parametrize(
    "status_code,expected_status_code",
    [
        (301, 301),
        (302, 302),
        (303, 303),
        (307, 307),
        (308, 308),
    ],
)
def test_redirect_dynamic_status_code(status_code: Optional[int], expected_status_code: int) -> None:
    @get("/")
    def handler() -> Redirect:
        return Redirect(path="/something-else", status_code=status_code)  # type: ignore[arg-type]

    with create_test_client(
        [handler],
    ) as client:
        res = client.get("/", follow_redirects=False)
        assert res.status_code == expected_status_code


@pytest.mark.parametrize("handler_status_code", [301, 307, None])
def test_redirect(handler_status_code: Optional[int]) -> None:
    @get("/", status_code=handler_status_code)
    def handler() -> Redirect:
        return Redirect(path="/something-else", status_code=handler_status_code)  # type: ignore[arg-type]

    with create_test_client([handler]) as client:
        res = client.get("/", follow_redirects=False)
        assert res.status_code == 302 if handler_status_code is None else handler_status_code



from typing import TYPE_CHECKING, Any, List

import pytest

from litestar import HttpMethod, Litestar, MediaType, get
from litestar.exceptions import ImproperlyConfiguredException
from litestar.static_files import StaticFilesConfig, create_static_files_router
from litestar.status_codes import HTTP_200_OK, HTTP_204_NO_CONTENT, HTTP_405_METHOD_NOT_ALLOWED
from litestar.testing import create_test_client

if TYPE_CHECKING:
    from pathlib import Path


@pytest.mark.parametrize("directories", [[], [""]])
@pytest.mark.parametrize("func", [StaticFilesConfig, create_static_files_router])
def test_config_validation_of_directories(func: Any, directories: List[str]) -> None:
    with pytest.raises(ImproperlyConfiguredException):
        func(path="/static", directories=directories)


@pytest.mark.parametrize("func", [StaticFilesConfig, create_static_files_router])
def test_config_validation_of_path(tmpdir: "Path", func: Any) -> None:
    path = tmpdir / "text.txt"
    path.write_text("content", "utf-8")

    with pytest.raises(ImproperlyConfiguredException):
        func(path="", directories=[tmpdir])

    with pytest.raises(ImproperlyConfiguredException):
        func(path="/{param:int}", directories=[tmpdir])


@pytest.mark.parametrize("func", [StaticFilesConfig, create_static_files_router])
def test_config_validation_of_file_system(tmpdir: "Path", func: Any) -> None:
    class FSWithoutOpen:
        def info(self) -> None:
            return

    with pytest.raises(ImproperlyConfiguredException):
        func(path="/static", directories=[tmpdir], file_system=FSWithoutOpen())

    class FSWithoutInfo:
        def open(self) -> None:
            return

    with pytest.raises(ImproperlyConfiguredException):
        func(path="/static", directories=[tmpdir], file_system=FSWithoutInfo())

    class ImplementedFS:
        def info(self) -> None:
            return

        def open(self) -> None:
            return

    assert func(path="/static", directories=[tmpdir], file_system=ImplementedFS())


def test_runtime_validation_of_static_path_and_path_parameter(tmpdir: "Path") -> None:
    path = tmpdir / "test.txt"
    path.write_text("content", "utf-8")

    @get("/static/{f:str}", media_type=MediaType.TEXT)
    def handler(f: str) -> str:
        return f

    with pytest.raises(ImproperlyConfiguredException):
        Litestar(
            route_handlers=[handler], static_files_config=[StaticFilesConfig(path="/static", directories=[tmpdir])]
        )


@pytest.mark.parametrize(
    "method, expected",
    (
        (HttpMethod.GET, HTTP_200_OK),
        (HttpMethod.HEAD, HTTP_200_OK),
        (HttpMethod.PUT, HTTP_405_METHOD_NOT_ALLOWED),
        (HttpMethod.PATCH, HTTP_405_METHOD_NOT_ALLOWED),
        (HttpMethod.POST, HTTP_405_METHOD_NOT_ALLOWED),
        (HttpMethod.DELETE, HTTP_405_METHOD_NOT_ALLOWED),
        (HttpMethod.OPTIONS, HTTP_405_METHOD_NOT_ALLOWED),
    ),
)
def test_runtime_validation_of_request_method_legacy_config(tmpdir: "Path", method: HttpMethod, expected: int) -> None:
    path = tmpdir / "test.txt"
    path.write_text("content", "utf-8")

    with create_test_client(
        [], static_files_config=[StaticFilesConfig(path="/static", directories=[tmpdir])]
    ) as client:
        response = client.request(method, "/static/test.txt")
        assert response.status_code == expected


@pytest.mark.parametrize(
    "method, expected",
    (
        (HttpMethod.GET, HTTP_200_OK),
        (HttpMethod.HEAD, HTTP_200_OK),
        (HttpMethod.OPTIONS, HTTP_204_NO_CONTENT),
        (HttpMethod.PUT, HTTP_405_METHOD_NOT_ALLOWED),
        (HttpMethod.PATCH, HTTP_405_METHOD_NOT_ALLOWED),
        (HttpMethod.POST, HTTP_405_METHOD_NOT_ALLOWED),
        (HttpMethod.DELETE, HTTP_405_METHOD_NOT_ALLOWED),
    ),
)
def test_runtime_validation_of_request_method_create_handler(tmpdir: "Path", method: HttpMethod, expected: int) -> None:
    path = tmpdir / "test.txt"
    path.write_text("content", "utf-8")

    with create_test_client(create_static_files_router(path="/static", directories=[tmpdir])) as client:
        response = client.request(method, "/static/test.txt")
        assert response.status_code == expected



from __future__ import annotations

from typing import Any, Generic, TypeVar
from urllib.parse import quote

from litestar import Response
from litestar.contrib.htmx._utils import HTMX_STOP_POLLING, get_headers
from litestar.contrib.htmx.types import (
    EventAfterType,
    HtmxHeaderType,
    LocationType,
    PushUrlType,
    ReSwapMethod,
    TriggerEventType,
)
from litestar.response import Template
from litestar.status_codes import HTTP_200_OK

__all__ = (
    "ClientRedirect",
    "ClientRefresh",
    "HTMXTemplate",
    "HXLocation",
    "HXStopPolling",
    "PushUrl",
    "ReplaceUrl",
    "Reswap",
    "Retarget",
    "TriggerEvent",
)


# HTMX defined HTTP status code.
# Response carrying this status code will ask client to stop Polling.
T = TypeVar("T")


class HXStopPolling(Response):
    """Stop HTMX client from Polling."""

    def __init__(self) -> None:
        """Initialize"""
        super().__init__(content=None)
        self.status_code = HTMX_STOP_POLLING


class ClientRedirect(Response):
    """HTMX Response class to support client side redirect."""

    def __init__(self, redirect_to: str) -> None:
        """Set status code to 200 (required by HTMX), and pass redirect url."""
        super().__init__(content=None, headers=get_headers(hx_headers=HtmxHeaderType(redirect=redirect_to)))
        del self.headers["Location"]


class ClientRefresh(Response):
    """Response to support HTMX client page refresh"""

    def __init__(self) -> None:
        """Set Status code to 200 and set headers."""
        super().__init__(content=None, headers=get_headers(hx_headers=HtmxHeaderType(refresh=True)))


class PushUrl(Generic[T], Response[T]):
    """Response to push new url into the history stack."""

    def __init__(self, content: T, push_url: PushUrlType, **kwargs: Any) -> None:
        """Initialize PushUrl."""
        super().__init__(
            content=content,
            status_code=HTTP_200_OK,
            headers=get_headers(hx_headers=HtmxHeaderType(push_url=push_url)),
            **kwargs,
        )


class ReplaceUrl(Generic[T], Response[T]):
    """Response to replace url in the Browser Location bar."""

    def __init__(self, content: T, replace_url: PushUrlType, **kwargs: Any) -> None:
        """Initialize ReplaceUrl."""
        super().__init__(
            content=content,
            status_code=HTTP_200_OK,
            headers=get_headers(hx_headers=HtmxHeaderType(replace_url=replace_url)),
            **kwargs,
        )


class Reswap(Generic[T], Response[T]):
    """Response to specify how the response will be swapped."""

    def __init__(
        self,
        content: T,
        method: ReSwapMethod,
        **kwargs: Any,
    ) -> None:
        """Initialize Reswap."""
        super().__init__(content=content, headers=get_headers(hx_headers=HtmxHeaderType(re_swap=method)), **kwargs)


class Retarget(Generic[T], Response[T]):
    """Response to target different element on the page."""

    def __init__(self, content: T, target: str, **kwargs: Any) -> None:
        """Initialize Retarget."""
        super().__init__(content=content, headers=get_headers(hx_headers=HtmxHeaderType(re_target=target)), **kwargs)


class TriggerEvent(Generic[T], Response[T]):
    """Trigger Client side event."""

    def __init__(
        self,
        content: T,
        name: str,
        after: EventAfterType,
        params: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize TriggerEvent."""
        event = TriggerEventType(name=name, params=params, after=after)
        headers = get_headers(hx_headers=HtmxHeaderType(trigger_event=event))
        super().__init__(content=content, headers=headers, **kwargs)


class HXLocation(Response):
    """Client side redirect without full page reload."""

    def __init__(
        self,
        redirect_to: str,
        source: str | None = None,
        event: str | None = None,
        target: str | None = None,
        swap: ReSwapMethod | None = None,
        hx_headers: dict[str, Any] | None = None,
        values: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize HXLocation, Set status code to 200 (required by HTMX),
        and pass redirect url.
        """
        super().__init__(
            content=None,
            headers={"Location": quote(redirect_to, safe="/#%[]=:;$&()+,!?*@'~")},
            **kwargs,
        )
        spec: dict[str, Any] = get_headers(
            hx_headers=HtmxHeaderType(
                location=LocationType(
                    path=str(self.headers.get("Location")),
                    source=source,
                    event=event,
                    target=target,
                    swap=swap,
                    values=values,
                    hx_headers=hx_headers,
                )
            )
        )
        del self.headers["Location"]
        self.headers.update(spec)


class HTMXTemplate(Template):
    """HTMX template wrapper"""

    def __init__(
        self,
        push_url: PushUrlType | None = None,
        re_swap: ReSwapMethod | None = None,
        re_target: str | None = None,
        trigger_event: str | None = None,
        params: dict[str, Any] | None = None,
        after: EventAfterType | None = None,
        **kwargs: Any,
    ) -> None:
        """Create HTMXTemplate response.

        Args:
            push_url: Either a string value specifying a URL to push to browser history or ``False`` to prevent HTMX client from
                pushing a url to browser history.
            re_swap: Method value to instruct HTMX which swapping method to use.
            re_target: Value for 'id of target element' to apply changes to.
            trigger_event: Event name to trigger.
            params: Dictionary of parameters if any required with trigger event parameter.
            after: Changes to apply after ``receive``, ``settle`` or ``swap`` event.
            **kwargs: Additional arguments to pass to ``Template``.
        """
        super().__init__(**kwargs)

        event: TriggerEventType | None = None
        if trigger_event:
            event = TriggerEventType(name=str(trigger_event), params=params, after=after)

        self.headers.update(
            get_headers(HtmxHeaderType(push_url=push_url, re_swap=re_swap, re_target=re_target, trigger_event=event))
        )



import enum
from pathlib import Path, PurePath, PureWindowsPath
from typing import Any, Callable, cast

import msgspec
import pytest
from pytest import FixtureRequest

from litestar import MediaType, Response
from litestar.exceptions import ImproperlyConfiguredException
from litestar.serialization import get_serializer
from tests.models import DataclassPersonFactory, MsgSpecStructPerson

person = DataclassPersonFactory.build()


class _TestEnum(enum.Enum):
    A = "alpha"
    B = "beta"


@pytest.fixture(params=[MediaType.JSON, MediaType.MESSAGEPACK])
def media_type(request: FixtureRequest) -> MediaType:
    return cast(MediaType, request.param)


DecodeMediaType = Callable[[Any], Any]


@pytest.fixture()
def decode_media_type(media_type: MediaType) -> DecodeMediaType:
    if media_type == MediaType.JSON:
        return msgspec.json.decode
    return msgspec.msgpack.decode


def test_dataclass(media_type: MediaType, decode_media_type: DecodeMediaType) -> None:
    encoded = Response(None).render(person, media_type=media_type)
    assert decode_media_type(encoded) == msgspec.to_builtins(person)


def test_struct(media_type: MediaType, decode_media_type: DecodeMediaType) -> None:
    encoded = Response(None).render(MsgSpecStructPerson(**msgspec.to_builtins(person)), media_type=media_type)
    assert decode_media_type(encoded) == msgspec.to_builtins(person)


@pytest.mark.parametrize("content", [{"value": 1}, [{"value": 1}]])
def test_dict(media_type: MediaType, decode_media_type: DecodeMediaType, content: Any) -> None:
    encoded = Response(None).render(content, media_type=media_type)
    assert decode_media_type(encoded) == content


def test_enum(media_type: MediaType, decode_media_type: DecodeMediaType) -> None:
    encoded = Response(None).render({"value": _TestEnum.A}, media_type=media_type)
    assert decode_media_type(encoded) == {"value": _TestEnum.A.value}


@pytest.mark.parametrize("path", [PurePath("/path/to/file"), Path("/path/to/file")])
def test_path(media_type: MediaType, decode_media_type: DecodeMediaType, path: Path) -> None:
    encoded = Response(None).render({"value": path}, media_type=media_type)
    expected = r"\path\to\file" if isinstance(path, PureWindowsPath) else "/path/to/file"
    assert decode_media_type(encoded) == {"value": expected}


@pytest.mark.parametrize(
    "content, response_type, media_type", [["abcdefg", str, MediaType.TEXT], ["<div/>", str, MediaType.HTML]]
)
def test_response_serialization_text_types(content: Any, response_type: Any, media_type: MediaType) -> None:
    assert Response(None).render(content, media_type=media_type, enc_hook=get_serializer({})) == content.encode("utf-8")


@pytest.mark.parametrize(
    "content, response_type, media_type, should_raise",
    [
        ["abcdefg", str, "text/custom", False],
        ["<xml/>", str, "application/unknown", False],
        [b"<xml/>", bytes, "application/unknown", False],
        [{"key": "value"}, dict, "application/unknown", True],
    ],
)
def test_response_validation_of_unknown_media_types(
    content: Any, response_type: Any, media_type: MediaType, should_raise: bool
) -> None:
    response = Response(None)
    if should_raise:
        with pytest.raises(ImproperlyConfiguredException):
            response.render(content, media_type=media_type)
    else:
        rendered = response.render(content, media_type=media_type)
        assert rendered == (content if isinstance(content, bytes) else content.encode("utf-8"))



from __future__ import annotations

import asyncio
from datetime import timedelta
from typing import AsyncGenerator, cast
from unittest.mock import AsyncMock, MagicMock

import pytest
from _pytest.fixtures import FixtureRequest
from redis.asyncio.client import Redis

from litestar.channels import ChannelsBackend
from litestar.channels.backends.asyncpg import AsyncPgChannelsBackend
from litestar.channels.backends.memory import MemoryChannelsBackend
from litestar.channels.backends.psycopg import PsycoPgChannelsBackend
from litestar.channels.backends.redis import RedisChannelsStreamBackend
from litestar.exceptions import ImproperlyConfiguredException
from litestar.utils.compat import async_next


@pytest.fixture(
    params=[
        pytest.param("redis_pub_sub_backend", id="redis:pubsub", marks=pytest.mark.xdist_group("redis")),
        pytest.param("redis_stream_backend", id="redis:stream", marks=pytest.mark.xdist_group("redis")),
        pytest.param("postgres_asyncpg_backend", id="postgres:asyncpg", marks=pytest.mark.xdist_group("postgres")),
        pytest.param("postgres_psycopg_backend", id="postgres:psycopg", marks=pytest.mark.xdist_group("postgres")),
        pytest.param("memory_backend", id="memory"),
    ]
)
def channels_backend_instance(request: FixtureRequest) -> ChannelsBackend:
    return cast(ChannelsBackend, request.getfixturevalue(request.param))


@pytest.fixture(
    params=[
        pytest.param(
            "redis_stream_backend_with_history", id="redis:stream+history", marks=pytest.mark.xdist_group("redis")
        ),
        pytest.param("memory_backend_with_history", id="memory+history"),
    ]
)
def channels_backend_instance_with_history(request: FixtureRequest) -> ChannelsBackend:
    return cast(ChannelsBackend, request.getfixturevalue(request.param))


@pytest.fixture()
async def channels_backend(channels_backend_instance: ChannelsBackend) -> AsyncGenerator[ChannelsBackend, None]:
    await channels_backend_instance.on_startup()
    yield channels_backend_instance
    await channels_backend_instance.on_shutdown()


@pytest.fixture()
async def channels_backend_with_history(
    channels_backend_instance_with_history: ChannelsBackend,
) -> AsyncGenerator[ChannelsBackend, None]:
    await channels_backend_instance_with_history.on_startup()
    yield channels_backend_instance_with_history
    await channels_backend_instance_with_history.on_shutdown()


@pytest.mark.parametrize("channels", [{"foo"}, {"foo", "bar"}])
async def test_pub_sub(channels_backend: ChannelsBackend, channels: set[str]) -> None:
    await channels_backend.subscribe(channels)
    await channels_backend.publish(b"something", channels)

    event_generator = channels_backend.stream_events()
    received = set()
    for _ in channels:
        received.add(await async_next(event_generator))
    assert received == {(c, b"something") for c in channels}


async def test_pub_sub_unsubscribe(channels_backend: ChannelsBackend) -> None:
    await channels_backend.subscribe(["foo", "bar"])
    await channels_backend.publish(b"something", ["foo"])

    event_generator = channels_backend.stream_events()
    await channels_backend.unsubscribe(["foo"])
    await channels_backend.publish(b"something", ["bar"])

    assert await asyncio.wait_for(async_next(event_generator), timeout=0.01) == ("bar", b"something")


async def test_pub_sub_no_subscriptions(channels_backend: ChannelsBackend) -> None:
    await channels_backend.publish(b"something", ["foo"])

    event_generator = channels_backend.stream_events()
    with pytest.raises((asyncio.TimeoutError, TimeoutError)):
        await asyncio.wait_for(async_next(event_generator), timeout=0.01)


@pytest.mark.flaky(reruns=5)  # this should not really happen but just in case, we retry
async def test_pub_sub_no_subscriptions_by_unsubscribes(channels_backend: ChannelsBackend) -> None:
    await channels_backend.subscribe(["foo", "bar"])
    await channels_backend.publish(b"something", ["foo"])

    event_generator = channels_backend.stream_events()
    await asyncio.wait_for(async_next(event_generator), timeout=0.01)
    await channels_backend.unsubscribe(["foo"])
    await channels_backend.publish(b"something", ["foo"])

    with pytest.raises((asyncio.TimeoutError, TimeoutError)):
        await asyncio.wait_for(async_next(event_generator), timeout=0.01)


async def test_pub_sub_shutdown_leftover_messages(channels_backend_instance: ChannelsBackend) -> None:
    await channels_backend_instance.on_startup()

    await channels_backend_instance.publish(b"something", {"foo"})

    await asyncio.wait_for(channels_backend_instance.on_shutdown(), timeout=0.1)


async def test_unsubscribe_without_subscription(channels_backend: ChannelsBackend) -> None:
    await channels_backend.unsubscribe(["foo"])


@pytest.mark.parametrize("history_limit,expected_history_length", [(None, 10), (1, 1), (5, 5), (10, 10)])
async def test_get_history(
    channels_backend_with_history: ChannelsBackend, history_limit: int | None, expected_history_length: int
) -> None:
    messages = [str(i).encode() for i in range(100)]
    for message in messages:
        await channels_backend_with_history.publish(message, {"something"})

    history = await channels_backend_with_history.get_history("something", history_limit)

    expected_messages = messages[-expected_history_length:]
    assert len(history) == expected_history_length
    assert history == expected_messages


async def test_discards_history_entries(channels_backend_with_history: ChannelsBackend) -> None:
    for _ in range(20):
        await channels_backend_with_history.publish(b"foo", {"bar"})

    assert len(await channels_backend_with_history.get_history("bar")) == 10


@pytest.mark.xdist_group("redis")
async def test_redis_streams_backend_flushall(redis_stream_backend: RedisChannelsStreamBackend) -> None:
    await redis_stream_backend.publish(b"something", ["foo", "bar", "baz"])

    result = await redis_stream_backend.flush_all()

    assert result == 3


@pytest.mark.flaky(reruns=5)  # this should not really happen but just in case, we retry
@pytest.mark.xdist_group("redis")
async def test_redis_stream_backend_expires(redis_client: Redis) -> None:
    backend = RedisChannelsStreamBackend(redis=redis_client, stream_ttl=timedelta(milliseconds=10), history=2)

    await backend.publish(b"something", ["foo"])
    await asyncio.sleep(0.1)
    await backend.publish(b"something", ["bar"])

    assert not await backend._redis.xrange(backend._make_key("foo"))
    assert await backend._redis.xrange(backend._make_key("bar"))


async def test_memory_publish_not_initialized_raises() -> None:
    backend = MemoryChannelsBackend()

    with pytest.raises(RuntimeError):
        await backend.publish(b"foo", ["something"])


@pytest.mark.xdist_group("postgres")
async def test_asyncpg_get_history(postgres_asyncpg_backend: AsyncPgChannelsBackend) -> None:
    with pytest.raises(NotImplementedError):
        await postgres_asyncpg_backend.get_history("something")


@pytest.mark.xdist_group("postgres")
async def test_psycopg_get_history(postgres_psycopg_backend: PsycoPgChannelsBackend) -> None:
    with pytest.raises(NotImplementedError):
        await postgres_psycopg_backend.get_history("something")


async def test_asyncpg_make_connection() -> None:
    make_connection = AsyncMock()

    backend = AsyncPgChannelsBackend(make_connection=make_connection)
    await backend.on_startup()

    make_connection.assert_awaited_once()


async def test_asyncpg_no_make_conn_or_dsn_passed_raises() -> None:
    with pytest.raises(ImproperlyConfiguredException):
        AsyncPgChannelsBackend()  # type: ignore[call-overload]


def test_asyncpg_listener_raises_on_non_string_payload() -> None:
    backend = AsyncPgChannelsBackend(make_connection=AsyncMock())
    with pytest.raises(RuntimeError):
        backend._listener(connection=MagicMock(), pid=1, payload=b"abc", channel="foo")


async def test_asyncpg_backend_publish_before_startup_raises() -> None:
    backend = AsyncPgChannelsBackend(make_connection=AsyncMock())

    with pytest.raises(RuntimeError):
        await backend.publish(b"foo", ["bar"])


async def test_asyncpg_backend_stream_before_startup_raises() -> None:
    backend = AsyncPgChannelsBackend(make_connection=AsyncMock())

    with pytest.raises(RuntimeError):
        await asyncio.wait_for(async_next(backend.stream_events()), timeout=0.01)


async def test_memory_backend_stream_before_startup_raises() -> None:
    backend = MemoryChannelsBackend()

    with pytest.raises(RuntimeError):
        await asyncio.wait_for(async_next(backend.stream_events()), timeout=0.01)



"""Plugin for creating and retrieving flash messages."""

from dataclasses import dataclass
from typing import Any, Mapping

from litestar.config.app import AppConfig
from litestar.connection import ASGIConnection
from litestar.contrib.minijinja import MiniJinjaTemplateEngine
from litestar.plugins import InitPluginProtocol
from litestar.template import TemplateConfig
from litestar.template.base import _get_request_from_context
from litestar.utils.scope.state import ScopeState


@dataclass
class FlashConfig:
    """Configuration for Flash messages."""

    template_config: TemplateConfig


class FlashPlugin(InitPluginProtocol):
    """Flash messages Plugin."""

    def __init__(self, config: FlashConfig):
        """Initialize the plugin.

        Args:
            config: Configuration for flash messages, including the template engine instance.
        """
        self.config = config

    def on_app_init(self, app_config: AppConfig) -> AppConfig:
        """Register the message callable on the template engine instance.

        Args:
            app_config: The application configuration.

        Returns:
            The application configuration with the message callable registered.
        """
        if isinstance(self.config.template_config.engine_instance, MiniJinjaTemplateEngine):
            from litestar.contrib.minijinja import _transform_state

            self.config.template_config.engine_instance.register_template_callable(
                "get_flashes", _transform_state(get_flashes)
            )
        else:
            self.config.template_config.engine_instance.register_template_callable("get_flashes", get_flashes)
        return app_config


def flash(connection: ASGIConnection, message: str, category: str) -> None:
    """Add a flash message to the request scope.

    Args:
        connection: The connection instance.
        message: The message to flash.
        category: The category of the message.
    """
    scope_state = ScopeState.from_scope(connection.scope)
    scope_state.flash_messages.append({"message": message, "category": category})


def get_flashes(context: Mapping[str, Any]) -> Any:
    """Get flash messages from the request scope, if any.

    Args:
        context: The context dictionary.

    Returns:
        The flash messages, if any.
    """
    scope_state = ScopeState.from_scope(_get_request_from_context(context).scope)
    return scope_state.flash_messages



from __future__ import annotations

import inspect
import multiprocessing
import os
import subprocess
import sys
from contextlib import AbstractContextManager, ExitStack, contextmanager
from typing import TYPE_CHECKING, Any, Iterator

import click
from click import Context, command, option
from rich.tree import Tree

from litestar.app import DEFAULT_OPENAPI_CONFIG
from litestar.cli._utils import (
    UVICORN_INSTALLED,
    LitestarEnv,
    console,
    create_ssl_files,
    isatty,
    remove_default_schema_routes,
    remove_routes_with_patterns,
    show_app_info,
    validate_ssl_file_paths,
)
from litestar.routes import ASGIRoute, HTTPRoute, WebSocketRoute
from litestar.utils.helpers import unwrap_partial

__all__ = ("info_command", "routes_command", "run_command")

if TYPE_CHECKING:
    from litestar import Litestar


@contextmanager
def _server_lifespan(app: Litestar) -> Iterator[None]:
    """Context manager handling the ASGI server lifespan.

    It will be entered just before the ASGI server is started through the CLI.
    """
    with ExitStack() as exit_stack:
        for manager in app._server_lifespan_managers:
            if not isinstance(manager, AbstractContextManager):
                manager = manager(app)  # type: ignore[assignment]
            exit_stack.enter_context(manager)  # type: ignore[arg-type]

        yield


def _convert_uvicorn_args(args: dict[str, Any]) -> list[str]:
    process_args = []
    for arg, value in args.items():
        if isinstance(value, bool):
            if value:
                process_args.append(f"--{arg}")
        elif isinstance(value, tuple):
            process_args.extend(f"--{arg}={item}" for item in value)
        else:
            process_args.append(f"--{arg}={value}")

    return process_args


def _run_uvicorn_in_subprocess(
    *,
    env: LitestarEnv,
    host: str | None,
    port: int | None,
    workers: int | None,
    reload: bool,
    reload_dirs: tuple[str, ...] | None,
    reload_include: tuple[str, ...] | None,
    reload_exclude: tuple[str, ...] | None,
    fd: int | None,
    uds: str | None,
    certfile_path: str | None,
    keyfile_path: str | None,
) -> None:
    process_args: dict[str, Any] = {
        "reload": reload,
        "host": host,
        "port": port,
        "workers": workers,
        "factory": env.is_app_factory,
    }
    if fd is not None:
        process_args["fd"] = fd
    if uds is not None:
        process_args["uds"] = uds
    if reload_dirs:
        process_args["reload-dir"] = reload_dirs
    if reload_include:
        process_args["reload-include"] = reload_include
    if reload_exclude:
        process_args["reload-exclude"] = reload_exclude
    if certfile_path is not None:
        process_args["ssl-certfile"] = certfile_path
    if keyfile_path is not None:
        process_args["ssl-keyfile"] = keyfile_path
    subprocess.run(
        [sys.executable, "-m", "uvicorn", env.app_path, *_convert_uvicorn_args(process_args)],  # noqa: S603
        check=True,
    )


class CommaSplittedPath(click.Path):
    """A Click Path that splits the input string by commas.

    .. versionadded:: 2.8.0
    """

    envvar_list_splitter = ","


@command(name="version")
@option("-s", "--short", help="Exclude release level and serial information", is_flag=True, default=False)
def version_command(short: bool) -> None:
    """Show the currently installed Litestar version."""
    from litestar import __version__

    click.echo(__version__.formatted(short=short))


@command(name="info")
def info_command(app: Litestar) -> None:
    """Show information about the detected Litestar app."""

    show_app_info(app)


@command(name="run")
@option("-r", "--reload", help="Reload server on changes", default=False, is_flag=True, envvar="LITESTAR_RELOAD")
@option(
    "-R",
    "--reload-dir",
    help="Directories to watch for file changes",
    type=CommaSplittedPath(),
    multiple=True,
    envvar="LITESTAR_RELOAD_DIRS",
)
@option(
    "-I",
    "--reload-include",
    help="Glob patterns for files to include when watching for file changes",
    type=CommaSplittedPath(),
    multiple=True,
    envvar="LITESTAR_RELOAD_INCLUDES",
)
@option(
    "-E",
    "--reload-exclude",
    help="Glob patterns for files to exclude when watching for file changes",
    type=CommaSplittedPath(),
    multiple=True,
    envvar="LITESTAR_RELOAD_EXCLUDES",
)
@option("-p", "--port", help="Serve under this port", type=int, default=8000, show_default=True, envvar="LITESTAR_PORT")
@option(
    "-W",
    "--wc",
    "--web-concurrency",
    help="The number of HTTP workers to launch",
    type=click.IntRange(min=1, max=multiprocessing.cpu_count() + 1),
    show_default=True,
    default=1,
    envvar=["LITESTAR_WEB_CONCURRENCY", "WEB_CONCURRENCY"],
)
@option("-H", "--host", help="Server under this host", default="127.0.0.1", show_default=True, envvar="LITESTAR_HOST")
@option(
    "-F",
    "--fd",
    "--file-descriptor",
    help="Bind to a socket from this file descriptor.",
    type=int,
    default=None,
    show_default=True,
    envvar="LITESTAR_FILE_DESCRIPTOR",
)
@option(
    "-U",
    "--uds",
    "--unix-domain-socket",
    help="Bind to a UNIX domain socket.",
    default=None,
    show_default=True,
    envvar="LITESTAR_UNIX_DOMAIN_SOCKET",
)
@option("-d", "--debug", help="Run app in debug mode", is_flag=True, envvar="LITESTAR_DEBUG")
@option("-P", "--pdb", "--use-pdb", help="Drop into PDB on an exception", is_flag=True, envvar="LITESTAR_PDB")
@option("--ssl-certfile", help="Location of the SSL cert file", default=None, envvar="LITESTAR_SSL_CERT_PATH")
@option("--ssl-keyfile", help="Location of the SSL key file", default=None, envvar="LITESTAR_SSL_KEY_PATH")
@option(
    "--create-self-signed-cert",
    help="If certificate and key are not found at specified locations, create a self-signed certificate and a key",
    is_flag=True,
    envvar="LITESTAR_CREATE_SELF_SIGNED_CERT",
)
def run_command(
    reload: bool,
    port: int,
    wc: int,
    host: str,
    fd: int | None,
    uds: str | None,
    debug: bool,
    reload_dir: tuple[str, ...],
    reload_include: tuple[str, ...],
    reload_exclude: tuple[str, ...],
    pdb: bool,
    ssl_certfile: str | None,
    ssl_keyfile: str | None,
    create_self_signed_cert: bool,
    ctx: Context,
) -> None:
    """Run a Litestar app; requires ``uvicorn``.

    The app can be either passed as a module path in the form of <module name>.<submodule>:<app instance or factory>,
    set as an environment variable LITESTAR_APP with the same format or automatically discovered from one of these
    canonical paths: app.py, asgi.py, application.py or app/__init__.py. When auto-discovering application factories,
    functions with the name ``create_app`` are considered, or functions that are annotated as returning a ``Litestar``
    instance.
    """

    if debug:
        os.environ["LITESTAR_DEBUG"] = "1"

    if pdb:
        os.environ["LITESTAR_PDB"] = "1"
    quiet_console = os.getenv("LITESTAR_QUIET_CONSOLE") or False
    if not UVICORN_INSTALLED:
        console.print(
            r"uvicorn is not installed. Please install the standard group, litestar\[standard], to use this command."
        )
        sys.exit(1)

    if callable(ctx.obj):
        ctx.obj = ctx.obj()
    else:
        if debug:
            ctx.obj.app.debug = True
        if pdb:
            ctx.obj.app.pdb_on_exception = True

    env: LitestarEnv = ctx.obj
    app = env.app

    reload = reload or bool(reload_dir) or bool(reload_include) or bool(reload_exclude)
    workers = wc

    certfile_path, keyfile_path = (
        create_ssl_files(ssl_certfile, ssl_keyfile, host)
        if create_self_signed_cert
        else validate_ssl_file_paths(ssl_certfile, ssl_keyfile)
    )

    if not quiet_console and isatty():
        console.rule("[yellow]Starting server process", align="left")
        show_app_info(app)
    with _server_lifespan(app):
        if workers == 1 and not reload:
            import uvicorn

            # A guard statement at the beginning of this function prevents uvicorn from being unbound
            # See "reportUnboundVariable in:
            # https://microsoft.github.io/pyright/#/configuration?id=type-check-diagnostics-settings
            uvicorn.run(  # pyright: ignore
                app=env.app_path,
                host=host,
                port=port,
                fd=fd,
                uds=uds,
                factory=env.is_app_factory,
                ssl_certfile=certfile_path,
                ssl_keyfile=keyfile_path,
            )
        else:
            # invoke uvicorn in a subprocess to be able to use the --reload flag. see
            # https://github.com/litestar-org/litestar/issues/1191 and https://github.com/encode/uvicorn/issues/1045
            if sys.gettrace() is not None:
                console.print(
                    "[yellow]Debugger detected. Breakpoints might not work correctly inside route handlers when running"
                    " with the --reload or --workers options[/]"
                )

            _run_uvicorn_in_subprocess(
                env=env,
                host=host,
                port=port,
                workers=workers,
                reload=reload,
                reload_dirs=reload_dir,
                reload_include=reload_include,
                reload_exclude=reload_exclude,
                fd=fd,
                uds=uds,
                certfile_path=certfile_path,
                keyfile_path=keyfile_path,
            )


@command(name="routes")
@option("--schema", help="Include schema routes", is_flag=True, default=False)
@option("--exclude", help="routes to exclude via regex", type=str, is_flag=False, multiple=True)
def routes_command(app: Litestar, exclude: tuple[str, ...], schema: bool) -> None:  # pragma: no cover
    """Display information about the application's routes."""

    sorted_routes = sorted(app.routes, key=lambda r: r.path)
    if not schema:
        openapi_config = app.openapi_config or DEFAULT_OPENAPI_CONFIG
        sorted_routes = remove_default_schema_routes(sorted_routes, openapi_config)
    if exclude is not None:
        sorted_routes = remove_routes_with_patterns(sorted_routes, exclude)

    console.print(_RouteTree(sorted_routes))


class _RouteTree(Tree):
    def __init__(self, routes: list[HTTPRoute | ASGIRoute | WebSocketRoute]) -> None:
        super().__init__("", hide_root=True)
        self._routes = routes
        self._build()

    def _build(self) -> None:
        for route in self._routes:
            if isinstance(route, HTTPRoute):
                self._handle_http_route(route)
            elif isinstance(route, WebSocketRoute):
                self._handle_websocket_route(route)
            else:
                self._handle_asgi_route(route)

    def _handle_asgi_like_route(self, route: ASGIRoute | WebSocketRoute, route_type: str) -> None:
        branch = self.add(f"[green]{route.path}[/green] ({route_type})")
        branch.add(f"[blue]{route.route_handler.name or route.route_handler.handler_name}[/blue]")

    def _handle_asgi_route(self, route: ASGIRoute) -> None:
        self._handle_asgi_like_route(route, route_type="ASGI")

    def _handle_websocket_route(self, route: WebSocketRoute) -> None:
        self._handle_asgi_like_route(route, route_type="WS")

    def _handle_http_route(self, route: HTTPRoute) -> None:
        branch = self.add(f"[green]{route.path}[/green] (HTTP)")
        for handler in route.route_handlers:
            handler_info = [
                f"[blue]{handler.name or handler.handler_name}[/blue]",
            ]

            if inspect.iscoroutinefunction(unwrap_partial(handler.fn)):
                handler_info.append("[magenta]async[/magenta]")
            else:
                handler_info.append("[yellow]sync[/yellow]")

            handler_info.append(f'[cyan]{", ".join(sorted(handler.http_methods))}[/cyan]')

            if len(handler.paths) > 1:
                for path in handler.paths:
                    branch.add(" ".join([f"[green]{path}[green]", *handler_info]))
            else:
                branch.add(" ".join(handler_info))



import pytest

from litestar import MediaType, get
from litestar.exceptions import LitestarWarning
from litestar.testing import create_test_client


def sync_handler() -> str:
    return "Hello World"


@pytest.mark.parametrize("sync_to_thread", [True, False])
def test_sync_to_thread(sync_to_thread: bool) -> None:
    handler = get("/", media_type=MediaType.TEXT, sync_to_thread=sync_to_thread)(sync_handler)

    with create_test_client(handler) as client:
        response = client.get("/")
        assert response.text == "Hello World"


@pytest.mark.usefixtures("enable_warn_implicit_sync_to_thread")
def test_sync_to_thread_not_set_warns() -> None:
    with pytest.warns(LitestarWarning, match="discouraged since synchronous callables"):
        get("/")(sync_handler)


@pytest.mark.parametrize("sync_to_thread", [True, False])
def test_async_callable_with_sync_to_thread_warns(sync_to_thread: bool) -> None:
    with pytest.warns(LitestarWarning, match="asynchronous callable"):

        @get(sync_to_thread=sync_to_thread)
        async def handler() -> None:
            pass



from __future__ import annotations

from dataclasses import MISSING
from inspect import Signature
from typing import Any, Final
from uuid import uuid4

from msgspec import UnsetType

from litestar.enums import MediaType
from litestar.types import Empty
from litestar.utils.deprecation import warn_deprecation

DEFAULT_ALLOWED_CORS_HEADERS: Final = {"Accept", "Accept-Language", "Content-Language", "Content-Type"}
DEFAULT_CHUNK_SIZE: Final = 1024 * 128  # 128KB
HTTP_DISCONNECT: Final = "http.disconnect"
HTTP_RESPONSE_BODY: Final = "http.response.body"
HTTP_RESPONSE_START: Final = "http.response.start"
ONE_MEGABYTE: Final = 1024 * 1024
OPENAPI_JSON_HANDLER_NAME: Final = f"{uuid4().hex}_litestar_openapi_json"
OPENAPI_NOT_INITIALIZED: Final = "Litestar has not been instantiated with OpenAPIConfig"
REDIRECT_STATUS_CODES: Final = {301, 302, 303, 307, 308}
REDIRECT_ALLOWED_MEDIA_TYPES: Final = {MediaType.TEXT, MediaType.HTML, MediaType.JSON}
RESERVED_KWARGS: Final = {"state", "headers", "cookies", "request", "socket", "data", "query", "scope", "body"}
SKIP_VALIDATION_NAMES: Final = {"request", "socket", "scope", "receive", "send"}
UNDEFINED_SENTINELS: Final = {Signature.empty, Empty, Ellipsis, MISSING, UnsetType}
WEBSOCKET_CLOSE: Final = "websocket.close"
WEBSOCKET_DISCONNECT: Final = "websocket.disconnect"

# deprecated constants
_SCOPE_STATE_CSRF_TOKEN_KEY = "csrf_token"  # noqa: S105  # possible hardcoded password
_SCOPE_STATE_DEPENDENCY_CACHE: Final = "dependency_cache"
_SCOPE_STATE_NAMESPACE: Final = "__litestar__"
_SCOPE_STATE_RESPONSE_COMPRESSED: Final = "response_compressed"
_SCOPE_STATE_DO_CACHE: Final = "do_cache"
_SCOPE_STATE_IS_CACHED: Final = "is_cached"

_deprecated_names = {
    "SCOPE_STATE_CSRF_TOKEN_KEY": _SCOPE_STATE_CSRF_TOKEN_KEY,
    "SCOPE_STATE_DEPENDENCY_CACHE": _SCOPE_STATE_DEPENDENCY_CACHE,
    "SCOPE_STATE_NAMESPACE": _SCOPE_STATE_NAMESPACE,
    "SCOPE_STATE_RESPONSE_COMPRESSED": _SCOPE_STATE_RESPONSE_COMPRESSED,
    "SCOPE_STATE_DO_CACHE": _SCOPE_STATE_DO_CACHE,
    "SCOPE_STATE_IS_CACHED": _SCOPE_STATE_IS_CACHED,
}


def __getattr__(name: str) -> Any:
    if name in _deprecated_names:
        warn_deprecation(
            deprecated_name=f"litestar.constants.{name}",
            version="2.4",
            kind="import",
            removal_in="3.0",
            info=f"'{name}' from 'litestar.constants' is deprecated and will be removed in 3.0. "
            "Direct access to Litestar scope state is not recommended.",
        )

        return globals()["_deprecated_names"][name]
    raise AttributeError(f"module {__name__} has no attribute {name}")  # pragma: no cover



from dataclasses import dataclass
from typing import Any, Dict, Literal, Optional
from uuid import UUID

from litestar import Request, get, post
from litestar.connection import ASGIConnection
from litestar.exceptions import NotAuthorizedException
from litestar.middleware.session.server_side import (
    ServerSideSessionBackend,
    ServerSideSessionConfig,
)
from litestar.security.session_auth import SessionAuth
from litestar.status_codes import HTTP_204_NO_CONTENT
from litestar.stores.memory import MemoryStore
from litestar.testing import create_test_client


@dataclass
class User:
    id: UUID
    name: str
    email: str


@dataclass
class UserLoginPayload:
    email: str
    password: str


MOCK_DB: Dict[str, User] = {}
memory_store = MemoryStore()


async def retrieve_user_handler(
    session: Dict[str, Any], connection: "ASGIConnection[Any, Any, Any, Any]"
) -> Optional[User]:
    return MOCK_DB.get(user_id) if (user_id := session.get("user_id")) else None


@post("/login")
async def login(data: UserLoginPayload, request: "Request[Any, Any, Any]") -> User:
    user_id_bytes = await memory_store.get(data.email)

    if not user_id_bytes:
        raise NotAuthorizedException

    user_id = user_id_bytes.decode("utf-8")
    request.set_session({"user_id": user_id})
    return MOCK_DB[user_id]


@get("/user", sync_to_thread=False)
def get_user(request: Request[User, Dict[Literal["user_id"], str], Any]) -> Any:
    return request.user


session_auth = SessionAuth[User, ServerSideSessionBackend](
    retrieve_user_handler=retrieve_user_handler,
    session_backend_config=ServerSideSessionConfig(),
    exclude=["/login", "/schema"],
)


def test_options_request_with_session_auth() -> None:
    with create_test_client(
        route_handlers=[login, get_user],
        on_app_init=[session_auth.on_app_init],
    ) as client:
        response = client.options(get_user.paths.pop())
        assert response.status_code == HTTP_204_NO_CONTENT



from typing import Type

import pytest
from hypothesis import given
from hypothesis import strategies as st
from starlette.exceptions import HTTPException as StarletteHTTPException

from litestar import get
from litestar.enums import MediaType
from litestar.exceptions import (
    HTTPException,
    ImproperlyConfiguredException,
    LitestarException,
    MissingDependencyException,
    ValidationException,
)
from litestar.middleware.exceptions.middleware import create_exception_response
from litestar.status_codes import HTTP_400_BAD_REQUEST, HTTP_500_INTERNAL_SERVER_ERROR
from litestar.testing import RequestFactory, create_test_client


class CustomLitestarException(LitestarException):
    detail = "Custom Exception"


class CustomHTTPException(HTTPException):
    detail = "Custom HTTP Exception"


@given(detail=st.text())
def test_litestar_exception_detail(detail: str) -> None:
    for result in LitestarException(detail=detail), LitestarException(detail):
        assert result.detail == detail


@given(detail=st.text())
def test_custom_litestar_exception_detail(detail: str) -> None:
    for result in CustomLitestarException(detail=detail), CustomLitestarException(detail):
        assert result.detail == (detail or "Custom Exception")


@given(detail=st.text())
@pytest.mark.parametrize("ex_type", [LitestarException, CustomLitestarException])
def test_litestar_exception_repr(ex_type: Type[LitestarException], detail: str) -> None:
    for result in ex_type(detail), ex_type(detail=detail):
        if result.detail:
            assert repr(result) == f"{result.__class__.__name__} - {result.detail}"
        else:
            assert repr(result) == result.__class__.__name__


@given(detail=st.text())
@pytest.mark.parametrize("ex_type", [LitestarException, CustomLitestarException])
def test_litestar_exception_str(ex_type: Type[LitestarException], detail: str) -> None:
    for result in ex_type(detail), ex_type(detail=detail):
        assert str(result) == result.detail.strip()

    result = ex_type(200, detail=detail)
    assert str(result) == f"200 {detail}".strip()


@given(detail=st.text())
def test_http_exception_detail(detail: str) -> None:
    for result in HTTPException(detail=detail), HTTPException(detail):
        assert result.detail == (detail or "Internal Server Error")


@given(detail=st.text())
def test_custom_http_exception_detail(detail: str) -> None:
    for result in CustomHTTPException(detail=detail), CustomHTTPException(detail):
        assert result.detail == (detail or "Custom HTTP Exception")


@given(status_code=st.integers(min_value=400, max_value=404), detail=st.text())
@pytest.mark.parametrize("ex_type", [HTTPException, CustomHTTPException])
def test_http_exception(ex_type: Type[HTTPException], status_code: int, detail: str) -> None:
    assert ex_type().status_code == HTTP_500_INTERNAL_SERVER_ERROR
    for result in ex_type(detail, status_code=status_code), ex_type(detail=detail, status_code=status_code):
        assert isinstance(result, LitestarException)
        assert repr(result) == f"{result.status_code} - {result.__class__.__name__} - {result.detail}"
        assert str(result) == f"{result.status_code}: {result.detail}".strip()


@given(detail=st.text())
def test_improperly_configured_exception(detail: str) -> None:
    result = ImproperlyConfiguredException(detail=detail)
    assert repr(result) == f"{HTTP_500_INTERNAL_SERVER_ERROR} - {result.__class__.__name__} - {result.detail}"
    assert isinstance(result, HTTPException)
    assert isinstance(result, ValueError)


def test_validation_exception() -> None:
    result = ValidationException()
    assert repr(result) == f"{HTTP_400_BAD_REQUEST} - {result.__class__.__name__} - {result.detail}"
    assert isinstance(result, HTTPException)
    assert isinstance(result, ValueError)


@pytest.mark.parametrize("media_type", [MediaType.JSON, MediaType.TEXT])
def test_create_exception_response_utility_litestar_http_exception(media_type: MediaType) -> None:
    exc = HTTPException(detail="litestar http exception", status_code=HTTP_400_BAD_REQUEST, extra=["any"])
    request = RequestFactory(handler_kwargs={"media_type": media_type}).get()
    response = create_exception_response(request=request, exc=exc)
    assert response.status_code == HTTP_400_BAD_REQUEST
    assert response.media_type == media_type
    if media_type == MediaType.JSON:
        assert response.content == {"status_code": 400, "detail": "litestar http exception", "extra": ["any"]}
    else:
        assert response.content == b'{"status_code":400,"detail":"litestar http exception","extra":["any"]}'


@pytest.mark.parametrize("media_type", [MediaType.JSON, MediaType.TEXT])
def test_create_exception_response_utility_starlette_http_exception(media_type: MediaType) -> None:
    @get("/", media_type=media_type)
    def handler() -> str:
        raise StarletteHTTPException(status_code=400)

    with create_test_client(handler) as client:
        response = client.get("/", headers={"Accept": media_type})
        assert response.json() == {"status_code": 400, "detail": "Bad Request"}


@pytest.mark.parametrize("media_type", [MediaType.JSON, MediaType.TEXT])
def test_create_exception_response_utility_non_http_exception(media_type: MediaType) -> None:
    exc = RuntimeError("yikes")
    request = RequestFactory(handler_kwargs={"media_type": media_type}).get()
    response = create_exception_response(request=request, exc=exc)
    assert response.status_code == HTTP_500_INTERNAL_SERVER_ERROR
    assert response.media_type == media_type
    if media_type == MediaType.JSON:
        assert response.content == {"status_code": 500, "detail": "Internal Server Error"}
    else:
        assert response.content == b'{"status_code":500,"detail":"Internal Server Error"}'


def test_missing_dependency_exception() -> None:
    exc = MissingDependencyException("some_package")
    expected = (
        "Package 'some_package' is not installed but required. You can install it by running 'pip install "
        "litestar[some_package]' to install litestar with the required extra or 'pip install some_package' to install "
        "the package separately"
    )
    assert str(exc) == expected


def test_missing_dependency_exception_differing_package_name() -> None:
    exc = MissingDependencyException("some_package", "install_via_this", "other-extra")
    expected = (
        "Package 'some_package' is not installed but required. You can install it by running 'pip install "
        "litestar[other-extra]' to install litestar with the required extra or 'pip install install_via_this' to "
        "install the package separately"
    )

    assert str(exc) == expected


@pytest.mark.parametrize("media_type", (MediaType.HTML, MediaType.JSON, MediaType.TEXT))
def test_default_exception_handling_of_internal_server_errors(media_type: MediaType) -> None:
    @get("/")
    def handler() -> None:
        raise ValueError("internal problem")

    with create_test_client(handler) as client:
        response = client.get("/", headers={"Accept": media_type})
        assert response.status_code == HTTP_500_INTERNAL_SERVER_ERROR
        if media_type == MediaType.HTML:
            assert response.text.startswith("<!doctype html>")
        elif media_type == MediaType.JSON:
            assert response.json().get("details").startswith("Traceback (most recent call last")
        else:
            assert response.text.startswith("Traceback (most recent call last")


def test_non_litestar_exception_with_status_code_is_500() -> None:
    # https://github.com/litestar-org/litestar/issues/3082
    class MyException(Exception):
        status_code: int = 400

    @get("/")
    def handler() -> None:
        raise MyException("hello")

    with create_test_client([handler]) as client:
        assert client.get("/").status_code == 500


def test_non_litestar_exception_with_detail_is_not_included() -> None:
    # https://github.com/litestar-org/litestar/issues/3082
    class MyException(Exception):
        status_code: int = 400
        detail: str = "hello"

    @get("/")
    def handler() -> None:
        raise MyException()

    with create_test_client([handler], debug=False) as client:
        assert client.get("/", headers={"Accept": MediaType.JSON}).json().get("detail") == "Internal Server Error"



from __future__ import annotations

from typing import (
    AsyncGenerator,
    Awaitable,
    Callable,
    Generic,
    Iterable,
    Iterator,
    TypeVar,
)

from typing_extensions import ParamSpec

from litestar.concurrency import sync_to_thread
from litestar.utils.predicates import is_async_callable

__all__ = ("ensure_async_callable", "AsyncIteratorWrapper", "AsyncCallable", "is_async_callable")


P = ParamSpec("P")
T = TypeVar("T")


def ensure_async_callable(fn: Callable[P, T]) -> Callable[P, Awaitable[T]]:
    """Ensure that ``fn`` is an asynchronous callable.
    If it is an asynchronous, return the original object, else wrap it in an
    ``AsyncCallable``
    """
    if is_async_callable(fn):
        return fn
    return AsyncCallable(fn)  # pyright: ignore


class AsyncCallable:
    """Wrap a given callable to be called in a thread pool using
    ``anyio.to_thread.run_sync``, keeping a reference to the original callable as
    :attr:`func`
    """

    def __init__(self, fn: Callable[P, T]) -> None:  # pyright: ignore
        self.func = fn

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> Awaitable[T]:  # pyright: ignore
        return sync_to_thread(self.func, *args, **kwargs)  # pyright: ignore


class AsyncIteratorWrapper(Generic[T]):
    """Asynchronous generator, wrapping an iterable or iterator."""

    __slots__ = ("iterator", "generator")

    def __init__(self, iterator: Iterator[T] | Iterable[T]) -> None:
        """Take a sync iterator or iterable and yields values from it asynchronously.

        Args:
            iterator: A sync iterator or iterable.
        """
        self.iterator = iterator if isinstance(iterator, Iterator) else iter(iterator)
        self.generator = self._async_generator()

    def _call_next(self) -> T:
        try:
            return next(self.iterator)
        except StopIteration as e:
            raise ValueError from e

    async def _async_generator(self) -> AsyncGenerator[T, None]:
        while True:
            try:
                yield await sync_to_thread(self._call_next)
            except ValueError:
                return

    def __aiter__(self) -> AsyncIteratorWrapper[T]:
        return self

    async def __anext__(self) -> T:
        return await self.generator.__anext__()



from itertools import permutations
from typing import TYPE_CHECKING, List, Mapping, Optional

import pytest

from litestar import get, route
from litestar.config.cors import CORSConfig
from litestar.status_codes import HTTP_204_NO_CONTENT, HTTP_400_BAD_REQUEST
from litestar.testing import create_test_client
from tests.helpers import RANDOM

if TYPE_CHECKING:
    from litestar.types import Method


@pytest.mark.parametrize(
    "http_methods",
    (list(perm) for perm in iter(permutations(["GET", "POST", "PATCH", "DELETE", "HEAD"], r=RANDOM.randrange(1, 6)))),
)
def test_regular_options_request(http_methods: List["Method"]) -> None:
    @route("/", http_method=http_methods)
    def handler() -> None:
        return None

    with create_test_client(handler, openapi_config=None) as client:
        response = client.options("/")
        assert response.status_code == HTTP_204_NO_CONTENT, response.text
        assert response.headers.get("Allow") == ", ".join(sorted({*http_methods, "OPTIONS"}))


def test_cors_options_request_without_origin_passes() -> None:
    @get("/")
    def handler() -> None:
        return None

    with create_test_client(handler, cors_config=CORSConfig(allow_origins=["http://testserver.local"])) as client:
        response = client.options("/")
        assert response.status_code == HTTP_204_NO_CONTENT
        assert response.headers.get("Allow") == "GET, OPTIONS"


def test_cors_options_request_with_correct_origin_passes() -> None:
    @get("/")
    def handler() -> None:
        return None

    with create_test_client(handler, cors_config=CORSConfig(allow_origins=["http://testserver.local"])) as client:
        response = client.options("/", headers={"Origin": "http://testserver.local"})
        assert response.status_code == HTTP_204_NO_CONTENT
        assert response.headers.get("Access-Control-Allow-Methods") == "DELETE, GET, HEAD, OPTIONS, PATCH, POST, PUT"
        assert response.headers.get("Access-Control-Allow-Origin") == "http://testserver.local"
        assert response.headers.get("Vary") == "Origin"


def test_cors_options_request_with_correct_origin_passes_with_allow_all_origins() -> None:
    @get("/")
    def handler() -> None:
        return None

    with create_test_client(handler, cors_config=CORSConfig(allow_origins=["*"])) as client:
        response = client.options("/", headers={"Origin": "http://testserver.local"})
        assert response.status_code == HTTP_204_NO_CONTENT
        assert response.headers.get("Access-Control-Allow-Methods") == "DELETE, GET, HEAD, OPTIONS, PATCH, POST, PUT"
        assert response.headers.get("Access-Control-Allow-Origin") == "*"
        assert "Vary" not in response.headers


def test_cors_options_request_with_wrong_origin_fails() -> None:
    @get("/")
    def handler() -> None:
        return None

    with create_test_client(handler, cors_config=CORSConfig(allow_origins=["http://testserver.local"])) as client:
        response = client.options("/", headers={"Origin": "https://moishe.zuchmir"})
        assert response.status_code == HTTP_400_BAD_REQUEST


@pytest.mark.parametrize(
    "allowed_origins, allowed_origin_regex, origin",
    (
        (["http://testserver.local", "https://moishe.zuchmir"], None, "https://moishe.zuchmir"),
        (["http://testserver.local", "https://moishe.zuchmir"], None, "http://testserver.local"),
        (["http://testserver.local", "https://moishe.*"], None, "https://moishe.zuchmir"),
        (["http://testserver.local", "https://moishe.*.abc.com"], None, "https://moishe.zuchmir.abc.com"),
        (["http://testserver.local", "https://moishe.*.*.com"], None, "https://moishe.zuchmir.zzz.com"),
        (["http://testserver.local"], "https://moishe.*.*.com", "https://moishe.zuchmir.zzz.com"),
        ([], "https://moishe.*.*.com", "https://moishe.zuchmir.zzz.com"),
    ),
)
def test_cors_options_request_with_different_domains_matches_regex(
    allowed_origins: List[str], allowed_origin_regex: Optional[str], origin: str
) -> None:
    @get("/")
    def handler() -> None:
        return None

    with create_test_client(
        handler, cors_config=CORSConfig(allow_origins=allowed_origins, allow_origin_regex=allowed_origin_regex)
    ) as client:
        response = client.options("/", headers={"Origin": origin})
        assert response.status_code == HTTP_204_NO_CONTENT


@pytest.mark.parametrize(
    "origin, allow_credentials",
    (("http://testserver.local", False), ("http://testserver.local", True), (None, False), (None, True)),
)
def test_cors_options_request_allow_credentials_header(origin: str, allow_credentials: bool) -> None:
    @get("/")
    def handler() -> None:
        return None

    with create_test_client(
        handler, cors_config=CORSConfig(allow_origins=["http://testserver.local"], allow_credentials=allow_credentials)
    ) as client:
        headers: Mapping[str, str] = {"Origin": origin} if origin else {}
        response = client.options("/", headers=headers)
        assert response.status_code == HTTP_204_NO_CONTENT

        if origin and allow_credentials:
            assert response.headers.get("Access-Control-Allow-Credentials") == str(allow_credentials).lower()
        else:
            assert "Access-Control-Allow-Credentials" not in response.headers


def test_cors_options_request_with_wrong_headers_fails() -> None:
    @get("/")
    def handler() -> None:
        return None

    with create_test_client(handler, cors_config=CORSConfig(allow_headers=["X-My-Header"])) as client:
        response = client.options(
            "/",
            headers={
                "Origin": "http://testserver.local",
                "Access-Control-Request-Headers": "X-My-Header, X-Another-Header",
            },
        )
        assert response.status_code == HTTP_400_BAD_REQUEST


def test_cors_options_request_with_correct_headers_passes() -> None:
    @get("/")
    def handler() -> None:
        return None

    with create_test_client(
        handler, cors_config=CORSConfig(allow_headers=["X-My-Header", "X-Another-Header"])
    ) as client:
        response = client.options(
            "/",
            headers={
                "Origin": "http://testserver.local",
                "Access-Control-Request-Headers": "X-My-Header, X-Another-Header",
            },
        )
        assert response.status_code == HTTP_204_NO_CONTENT


def test_requested_headers_are_reflected_back_when_allow_all_headers() -> None:
    @get("/")
    def handler() -> None:
        return None

    with create_test_client(handler, cors_config=CORSConfig(allow_headers=["*"])) as client:
        response = client.options(
            "/",
            headers={
                "Origin": "http://testserver.local",
                "Access-Control-Request-Headers": "X-My-Header, X-Another-Header",
            },
        )
        assert response.status_code == HTTP_204_NO_CONTENT
        assert (
            response.headers.get("Access-Control-Allow-Headers")
            == "Accept, Accept-Language, Content-Language, Content-Type, X-Another-Header, X-My-Header"
        )


def test_cors_options_request_fails_if_request_method_not_allowed() -> None:
    @get("/")
    def handler() -> None:
        return None

    with create_test_client(handler, cors_config=CORSConfig(allow_methods=["GET"])) as client:
        response = client.options(
            "/",
            headers={"Origin": "http://testserver.local", "Access-Control-Request-Method": "POST"},
        )
        assert response.status_code == HTTP_400_BAD_REQUEST


def test_cors_options_request_succeeds_if_request_method_not_specified() -> None:
    @get("/")
    def handler() -> None:
        return None

    with create_test_client(handler, cors_config=CORSConfig(allow_methods=["GET"])) as client:
        response = client.options(
            "/",
            headers={
                "Origin": "http://testserver.local",
            },
        )
        assert response.status_code == HTTP_204_NO_CONTENT
        assert response.headers.get("Access-Control-Allow-Methods") == "GET"



"""DTO domain types."""

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING

from litestar.exceptions import ImproperlyConfiguredException

if TYPE_CHECKING:
    from typing import Any, Literal, Mapping

    from litestar.typing import FieldDefinition

__all__ = (
    "DTO_FIELD_META_KEY",
    "DTOField",
    "Mark",
    "dto_field",
    "extract_dto_field",
)

DTO_FIELD_META_KEY = "__dto__"


class Mark(str, Enum):
    """For marking field definitions on domain models."""

    READ_ONLY = "read-only"
    """To mark a field that can be read, but not updated by clients."""
    WRITE_ONLY = "write-only"
    """To mark a field that can be written to, but not read by clients."""
    PRIVATE = "private"
    """To mark a field that can neither be read or updated by clients."""


@dataclass
class DTOField:
    """For configuring DTO behavior on model fields."""

    mark: Mark | Literal["read-only", "write-only", "private"] | None = None
    """Mark the field as read-only, or private."""


def dto_field(mark: Literal["read-only", "write-only", "private"] | Mark) -> dict[str, DTOField]:
    """Create a field metadata mapping.

    Args:
        mark: A DTO mark for the field, e.g., "read-only".

    Returns:
        A dict for setting as field metadata, such as the dataclass "metadata" field key, or the SQLAlchemy "info"
        field.

        Marking a field automates its inclusion/exclusion from DTO field definitions, depending on the DTO's purpose.
    """
    return {DTO_FIELD_META_KEY: DTOField(mark=Mark(mark))}


def extract_dto_field(field_definition: FieldDefinition, field_info_mapping: Mapping[str, Any]) -> DTOField:
    """Extract ``DTOField`` instance for a model field.

    Supports ``DTOField`` to bet set via ``Annotated`` or via a field info/metadata mapping.

    E.g., ``Annotated[str, DTOField(mark="read-only")]`` or ``info=dto_field(mark="read-only")``.

    If a value is found in ``field_info_mapping``, it is prioritized over the field definition's metadata.

    Args:
        field_definition: A field definition.
        field_info_mapping: A field metadata/info attribute mapping, e.g., SQLAlchemy's ``info`` attribute,
          or dataclasses ``metadata`` attribute.

    Returns:
        DTO field info, if any.
    """
    if inst := field_info_mapping.get(DTO_FIELD_META_KEY):
        if not isinstance(inst, DTOField):
            raise ImproperlyConfiguredException(f"DTO field info must be an instance of DTOField, got '{inst}'")
        return inst

    return next((f for f in field_definition.metadata if isinstance(f, DTOField)), DTOField())



from __future__ import annotations

from typing import TYPE_CHECKING

from litestar.static_files import StaticFilesConfig
from litestar.status_codes import HTTP_200_OK, HTTP_404_NOT_FOUND
from litestar.testing import create_test_client
from tests.unit.test_static_files.conftest import MakeConfig

if TYPE_CHECKING:
    from pathlib import Path

    from litestar.types import FileSystemProtocol


def test_staticfiles_is_html_mode(tmpdir: Path, file_system: FileSystemProtocol, make_config: MakeConfig) -> None:
    path = tmpdir / "index.html"
    path.write_text("content", "utf-8")
    static_files_config, handlers = make_config(
        StaticFilesConfig(path="/static", directories=[tmpdir], html_mode=True, file_system=file_system)
    )

    with create_test_client(handlers, static_files_config=static_files_config) as client:
        response = client.get("/static")
        assert response.status_code == HTTP_200_OK
        assert response.text == "content"
        assert response.headers["content-type"] == "text/html; charset=utf-8"
        assert response.headers["content-disposition"].startswith("inline")


def test_staticfiles_is_html_mode_serves_404_when_present(
    tmpdir: Path, file_system: FileSystemProtocol, make_config: MakeConfig
) -> None:
    path = tmpdir / "404.html"
    path.write_text("not found", "utf-8")
    static_files_config, handlers = make_config(
        StaticFilesConfig(path="/static", directories=[tmpdir], html_mode=True, file_system=file_system)
    )
    with create_test_client(handlers, static_files_config=static_files_config) as client:
        response = client.get("/static")
        assert response.status_code == HTTP_404_NOT_FOUND
        assert response.text == "not found"
        assert response.headers["content-type"] == "text/html; charset=utf-8"


def test_staticfiles_is_html_mode_raises_exception_when_no_404_html_is_present(
    tmpdir: Path, file_system: FileSystemProtocol, make_config: MakeConfig
) -> None:
    static_files_config, handlers = make_config(
        StaticFilesConfig(path="/static", directories=[tmpdir], html_mode=True, file_system=file_system)
    )
    with create_test_client(handlers, static_files_config=static_files_config) as client:
        response = client.get("/static")
        assert response.status_code == HTTP_404_NOT_FOUND
        assert response.json() == {"status_code": 404, "detail": "no file or directory match the path . was found"}



from __future__ import annotations

from copy import copy
from typing import Any

import pytest

from litestar.datastructures import State
from litestar.datastructures.state import ImmutableState


@pytest.mark.parametrize("state_class", (ImmutableState, State))
def test_state_immutable_mapping(state_class: type[ImmutableState]) -> None:
    state_dict = {"first": 1, "second": 2, "third": 3}
    state = state_class(state_dict, deep_copy=True)
    assert len(state) == 3
    assert "first" in state
    assert state["first"] == 1
    assert list(state.items()) == [("first", 1), ("second", 2), ("third", 3)]
    assert state
    assert isinstance(state.mutable_copy(), State)
    del state_dict["first"]
    assert "first" in state


@pytest.mark.parametrize(
    "zero_object", (ImmutableState({"first": 1}), State({"first": 1}), {"first": 1}, [("first", 1)])
)
def test_state_init(zero_object: Any) -> None:
    state = ImmutableState(zero_object)
    assert state.first


@pytest.mark.parametrize("zero_object", (ImmutableState({}), State(), {}, [], None))
def test_state_mapping(zero_object: Any) -> None:
    state = State(zero_object)
    assert not state
    state["first"] = "first"
    state["second"] = "second"
    assert state.first == "first"
    assert state["second"] == "second"
    del state["first"]
    del state.second
    assert "first" not in state
    assert "second" not in state
    assert isinstance(state.immutable_copy(), ImmutableState)


def test_state_attributes() -> None:
    state_dict = {"first": 1, "second": 2, "third": 3}
    state = State(state_dict)
    assert state.first == 1
    del state.first
    with pytest.raises(AttributeError):
        assert state.first
    state.fourth = 4
    assert state.fourth == 4
    with pytest.raises(AttributeError):
        del state.first


def test_state_dict() -> None:
    state_dict = {"first": 1, "second": 2, "third": 3}
    state = State(state_dict)
    assert state.dict() == state_dict


def test_state_copy() -> None:
    state_dict = {"key": {"inner": 1}}
    state = State(state_dict)
    copy = state.copy()
    del state.key
    assert copy.key


def test_state_copy_deep_copy_false() -> None:
    state = State({}, deep_copy=False)
    assert state.copy()._deep_copy is False


def test_unpicklable_deep_copy_false() -> None:
    # a module cannot be deep copied
    import typing

    state = ImmutableState({"module": typing}, deep_copy=False)
    copy(state)
    ImmutableState.validate(state)



from __future__ import annotations

from datetime import timedelta
from typing import TYPE_CHECKING, cast

from redis.asyncio import Redis
from redis.asyncio.connection import ConnectionPool

from litestar.exceptions import ImproperlyConfiguredException
from litestar.types import Empty, EmptyType
from litestar.utils.empty import value_or_default

from .base import NamespacedStore

__all__ = ("RedisStore",)

if TYPE_CHECKING:
    from types import TracebackType


class RedisStore(NamespacedStore):
    """Redis based, thread and process safe asynchronous key/value store."""

    __slots__ = (
        "_delete_all_script",
        "_get_and_renew_script",
        "_redis",
        "handle_client_shutdown",
    )

    def __init__(
        self, redis: Redis, namespace: str | None | EmptyType = Empty, handle_client_shutdown: bool = False
    ) -> None:
        """Initialize :class:`RedisStore`

        Args:
            redis: An :class:`redis.asyncio.Redis` instance
            namespace: A key prefix to simulate a namespace in redis. If not given,
                defaults to ``LITESTAR``. Namespacing can be explicitly disabled by passing
                ``None``. This will make :meth:`.delete_all` unavailable.
            handle_client_shutdown: If ``True``, handle the shutdown of the `redis` instance automatically during the store's lifespan. Should be set to `True` unless the shutdown is handled externally
        """
        self._redis = redis
        self.namespace: str | None = value_or_default(namespace, "LITESTAR")
        self.handle_client_shutdown = handle_client_shutdown

        # script to get and renew a key in one atomic step
        self._get_and_renew_script = self._redis.register_script(
            b"""
        local key = KEYS[1]
        local renew = tonumber(ARGV[1])

        local data = redis.call('GET', key)
        local ttl = redis.call('TTL', key)

        if ttl > 0 then
            redis.call('EXPIRE', key, renew)
        end

        return data
        """
        )

        # script to delete all keys in the namespace
        self._delete_all_script = self._redis.register_script(
            b"""
        local cursor = 0

        repeat
            local result = redis.call('SCAN', cursor, 'MATCH', ARGV[1])
            for _,key in ipairs(result[2]) do
                redis.call('UNLINK', key)
            end
            cursor = tonumber(result[1])
        until cursor == 0
        """
        )

    async def _shutdown(self) -> None:
        if self.handle_client_shutdown:
            await self._redis.aclose(close_connection_pool=True)  # type: ignore[attr-defined]

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self._shutdown()

    @classmethod
    def with_client(
        cls,
        url: str = "redis://localhost:6379",
        *,
        db: int | None = None,
        port: int | None = None,
        username: str | None = None,
        password: str | None = None,
        namespace: str | None | EmptyType = Empty,
    ) -> RedisStore:
        """Initialize a :class:`RedisStore` instance with a new class:`redis.asyncio.Redis` instance.

        Args:
            url: Redis URL to connect to
            db: Redis database to use
            port: Redis port to use
            username: Redis username to use
            password: Redis password to use
            namespace: Virtual key namespace to use
        """
        pool = ConnectionPool.from_url(
            url=url,
            db=db,
            decode_responses=False,
            port=port,
            username=username,
            password=password,
        )
        return cls(
            redis=Redis(connection_pool=pool),
            namespace=namespace,
            handle_client_shutdown=True,
        )

    def with_namespace(self, namespace: str) -> RedisStore:
        """Return a new :class:`RedisStore` with a nested virtual key namespace.
        The current instances namespace will serve as a prefix for the namespace, so it
        can be considered the parent namespace.
        """
        return type(self)(
            redis=self._redis,
            namespace=f"{self.namespace}_{namespace}" if self.namespace else namespace,
            handle_client_shutdown=self.handle_client_shutdown,
        )

    def _make_key(self, key: str) -> str:
        prefix = f"{self.namespace}:" if self.namespace else ""
        return prefix + key

    async def set(self, key: str, value: str | bytes, expires_in: int | timedelta | None = None) -> None:
        """Set a value.

        Args:
            key: Key to associate the value with
            value: Value to store
            expires_in: Time in seconds before the key is considered expired

        Returns:
            ``None``
        """
        if isinstance(value, str):
            value = value.encode("utf-8")
        await self._redis.set(self._make_key(key), value, ex=expires_in)

    async def get(self, key: str, renew_for: int | timedelta | None = None) -> bytes | None:
        """Get a value.

        Args:
            key: Key associated with the value
            renew_for: If given and the value had an initial expiry time set, renew the
                expiry time for ``renew_for`` seconds. If the value has not been set
                with an expiry time this is a no-op. Atomicity of this step is guaranteed
                by using a lua script to execute fetch and renewal. If ``renew_for`` is
                not given, the script will be bypassed so no overhead will occur

        Returns:
            The value associated with ``key`` if it exists and is not expired, else
            ``None``
        """
        key = self._make_key(key)
        if renew_for:
            if isinstance(renew_for, timedelta):
                renew_for = renew_for.seconds
            data = await self._get_and_renew_script(keys=[key], args=[renew_for])
            return cast("bytes | None", data)
        return await self._redis.get(key)

    async def delete(self, key: str) -> None:
        """Delete a value.

        If no such key exists, this is a no-op.

        Args:
            key: Key of the value to delete
        """
        await self._redis.delete(self._make_key(key))

    async def delete_all(self) -> None:
        """Delete all stored values in the virtual key namespace.

        Raises:
            ImproperlyConfiguredException: If no namespace was configured
        """
        if not self.namespace:
            raise ImproperlyConfiguredException("Cannot perform delete operation: No namespace configured")

        await self._delete_all_script(keys=[], args=[f"{self.namespace}*:*"])

    async def exists(self, key: str) -> bool:
        """Check if a given ``key`` exists."""
        return await self._redis.exists(self._make_key(key)) == 1

    async def expires_in(self, key: str) -> int | None:
        """Get the time in seconds ``key`` expires in. If no such ``key`` exists or no
        expiry time was set, return ``None``.
        """
        ttl = await self._redis.ttl(self._make_key(key))
        return None if ttl == -2 else ttl



from __future__ import annotations

from typing import Any, ClassVar, Collection, Iterable, Protocol, TypeVar, runtime_checkable

__all__ = (
    "DataclassProtocol",
    "InstantiableCollection",
    "Logger",
)


class Logger(Protocol):
    """Logger protocol."""

    def debug(self, event: str, *args: Any, **kwargs: Any) -> Any:
        """Output a log message at 'DEBUG' level.

        Args:
             event: Log message.
             *args: Any args.
             **kwargs: Any kwargs.
        """

    def info(self, event: str, *args: Any, **kwargs: Any) -> Any:
        """Output a log message at 'INFO' level.

        Args:
             event: Log message.
             *args: Any args.
             **kwargs: Any kwargs.
        """

    def warning(self, event: str, *args: Any, **kwargs: Any) -> Any:
        """Output a log message at 'WARNING' level.

        Args:
             event: Log message.
             *args: Any args.
             **kwargs: Any kwargs.
        """

    def warn(self, event: str, *args: Any, **kwargs: Any) -> Any:
        """Output a log message at 'WARN' level.

        Args:
             event: Log message.
             *args: Any args.
             **kwargs: Any kwargs.
        """

    def error(self, event: str, *args: Any, **kwargs: Any) -> Any:
        """Output a log message at 'ERROR' level.

        Args:
             event: Log message.
             *args: Any args.
             **kwargs: Any kwargs.
        """

    def fatal(self, event: str, *args: Any, **kwargs: Any) -> Any:
        """Output a log message at 'FATAL' level.

        Args:
             event: Log message.
             *args: Any args.
             **kwargs: Any kwargs.
        """

    def exception(self, event: str, *args: Any, **kwargs: Any) -> Any:
        """Log a message with level 'ERROR' on this logger. The arguments are interpreted as for debug(). Exception info
        is added to the logging message.

        Args:
             event: Log message.
             *args: Any args.
             **kwargs: Any kwargs.
        """

    def critical(self, event: str, *args: Any, **kwargs: Any) -> Any:
        """Output a log message at 'INFO' level.

        Args:
             event: Log message.
             *args: Any args.
             **kwargs: Any kwargs.
        """

    def setLevel(self, level: int) -> None:  # noqa: N802
        """Set the log level

        Args:
            level: Log level to set as an integer

        Returns:
            None
        """


@runtime_checkable
class DataclassProtocol(Protocol):
    """Protocol for instance checking dataclasses"""

    __dataclass_fields__: ClassVar[dict[str, Any]]


T_co = TypeVar("T_co", covariant=True)


@runtime_checkable
class InstantiableCollection(Collection[T_co], Protocol[T_co]):  # pyright: ignore
    """A protocol for instantiable collection types."""

    def __init__(self, iterable: Iterable[T_co], /) -> None: ...



from docs.examples.security.using_session_auth import app

from litestar.status_codes import HTTP_200_OK, HTTP_201_CREATED, HTTP_401_UNAUTHORIZED
from litestar.testing import TestClient


def test_using_session_auth_signup_flow() -> None:
    with TestClient(app) as client:
        response = client.get("/user")
        assert response.status_code == HTTP_401_UNAUTHORIZED
        response = client.post(
            "/signup", json={"name": "Moishe Zuchmir", "email": "moishe@zuchmir.com", "password": "abcd12345"}
        )
        assert response.status_code == HTTP_201_CREATED
        response = client.get("/user")
        assert response.status_code == HTTP_200_OK


def test_using_session_auth_login_flow() -> None:
    with TestClient(app) as client:
        response = client.post("/login", json={"email": "ludwig@zuchmir.com", "password": "abcd12345"})
        assert response.status_code == HTTP_401_UNAUTHORIZED
        response = client.post(
            "/signup", json={"name": "ludwig Zuchmir", "email": "ludwig@zuchmir.com", "password": "abcd12345"}
        )
        assert response.status_code == HTTP_201_CREATED
        response = client.post("/login", json={"email": "ludwig@zuchmir.com", "password": "abcd12345"})
        assert response.status_code == HTTP_201_CREATED
        response = client.get("/user")
        assert response.status_code == HTTP_200_OK



from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import datetime, timedelta, timezone
from typing import TYPE_CHECKING, Optional

from msgspec import Struct
from msgspec.msgpack import decode as msgpack_decode
from msgspec.msgpack import encode as msgpack_encode

if TYPE_CHECKING:
    from types import TracebackType

    from typing_extensions import Self


__all__ = ("Store", "NamespacedStore", "StorageObject")


class Store(ABC):
    """Thread and process safe asynchronous key/value store."""

    __slots__ = ()

    @abstractmethod
    async def set(self, key: str, value: str | bytes, expires_in: int | timedelta | None = None) -> None:
        """Set a value.

        Args:
            key: Key to associate the value with
            value: Value to store
            expires_in: Time in seconds before the key is considered expired

        Returns:
            ``None``
        """
        raise NotImplementedError

    @abstractmethod
    async def get(self, key: str, renew_for: int | timedelta | None = None) -> bytes | None:
        """Get a value.

        Args:
            key: Key associated with the value
            renew_for: If given and the value had an initial expiry time set, renew the
                expiry time for ``renew_for`` seconds. If the value has not been set
                with an expiry time this is a no-op

        Returns:
            The value associated with ``key`` if it exists and is not expired, else
            ``None``
        """
        raise NotImplementedError

    @abstractmethod
    async def delete(self, key: str) -> None:
        """Delete a value.

        If no such key exists, this is a no-op.

        Args:
            key: Key of the value to delete
        """
        raise NotImplementedError

    @abstractmethod
    async def delete_all(self) -> None:
        """Delete all stored values."""
        raise NotImplementedError

    @abstractmethod
    async def exists(self, key: str) -> bool:
        """Check if a given ``key`` exists."""
        raise NotImplementedError

    @abstractmethod
    async def expires_in(self, key: str) -> int | None:
        """Get the time in seconds ``key`` expires in. If no such ``key`` exists or no
        expiry time was set, return ``None``.
        """
        raise NotImplementedError

    async def __aenter__(self) -> None:  # noqa: B027
        pass

    async def __aexit__(  # noqa: B027
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        pass


class NamespacedStore(Store):
    """A subclass of :class:`Store`, offering hierarchical namespacing.

    Bulk actions on a parent namespace should affect all child namespaces, whereas other operations on all namespaces
    should be isolated.
    """

    __slots__ = ("namespace",)

    @abstractmethod
    def with_namespace(self, namespace: str) -> Self:
        """Return a new instance of :class:`NamespacedStore`, which exists in a child namespace of the current namespace.
        Bulk actions on the parent namespace should affect all child namespaces, whereas other operations on all
        namespaces should be isolated.
        """


class StorageObject(Struct):
    """:class:`msgspec.Struct` to store serialized data alongside with their expiry time."""

    expires_at: Optional[datetime]  # noqa: UP007
    data: bytes

    @classmethod
    def new(cls, data: bytes, expires_in: int | timedelta | None) -> StorageObject:
        """Construct a new :class:`StorageObject` instance."""
        if expires_in is not None and not isinstance(expires_in, timedelta):
            expires_in = timedelta(seconds=expires_in)
        return cls(
            data=data,
            expires_at=(datetime.now(tz=timezone.utc) + expires_in) if expires_in else None,
        )

    @property
    def expired(self) -> bool:
        """Return if the :class:`StorageObject` is expired"""
        return self.expires_at is not None and datetime.now(tz=timezone.utc) >= self.expires_at

    @property
    def expires_in(self) -> int:
        """Return the expiry time of this ``StorageObject`` in seconds. If no expiry time
        was set, return ``-1``.
        """
        if self.expires_at:
            return int(self.expires_at.timestamp() - datetime.now(tz=timezone.utc).timestamp())
        return -1

    def to_bytes(self) -> bytes:
        """Encode the instance to bytes"""
        return msgpack_encode(self)

    @classmethod
    def from_bytes(cls, raw: bytes) -> StorageObject:
        """Load a previously encoded with :meth:`StorageObject.to_bytes`"""
        return msgpack_decode(raw, type=cls)



import datetime
from decimal import Decimal
from typing import Any, Generic, Optional, Type, TypeVar, Union

import pydantic as pydantic_v2
import pytest
from pydantic import v1 as pydantic_v1
from pydantic.v1.generics import GenericModel
from typing_extensions import Annotated

from litestar._openapi.schema_generation import SchemaCreator
from litestar.contrib.pydantic.pydantic_schema_plugin import PydanticSchemaPlugin
from litestar.openapi.spec import OpenAPIType
from litestar.openapi.spec.schema import Schema
from litestar.typing import FieldDefinition
from litestar.utils.helpers import get_name
from tests.helpers import get_schema_for_field_definition

T = TypeVar("T")


class PydanticV1Generic(GenericModel, Generic[T]):
    foo: T
    optional_foo: Optional[T]
    annotated_foo: Annotated[T, object()]


class PydanticV2Generic(pydantic_v2.BaseModel, Generic[T]):
    foo: T
    optional_foo: Optional[T]
    annotated_foo: Annotated[T, object()]


@pytest.mark.parametrize("model", [PydanticV1Generic, PydanticV2Generic])
def test_schema_generation_with_generic_classes(model: Type[Union[PydanticV1Generic, PydanticV2Generic]]) -> None:
    cls = model[int]  # type: ignore[index]
    field_definition = FieldDefinition.from_kwarg(name=get_name(cls), annotation=cls)
    properties = get_schema_for_field_definition(field_definition, plugins=[PydanticSchemaPlugin()]).properties
    expected_foo_schema = Schema(type=OpenAPIType.INTEGER)
    expected_optional_foo_schema = Schema(one_of=[Schema(type=OpenAPIType.NULL), Schema(type=OpenAPIType.INTEGER)])

    assert properties
    assert properties["foo"] == expected_foo_schema
    assert properties["annotated_foo"] == expected_foo_schema
    assert properties["optional_foo"] == expected_optional_foo_schema


@pytest.mark.parametrize(
    "constrained",
    [
        pydantic_v1.constr(regex="^[a-zA-Z]$"),
        pydantic_v1.conlist(int, min_items=1),
        pydantic_v1.conset(int, min_items=1),
        pydantic_v1.conint(gt=10, lt=100),
        pydantic_v1.confloat(gt=10, lt=100),
        pydantic_v1.condecimal(gt=Decimal("10")),
        pydantic_v1.condate(gt=datetime.date.today()),
        pydantic_v2.constr(pattern="^[a-zA-Z]$"),
        pydantic_v2.conlist(int, min_length=1),
        pydantic_v2.conset(int, min_length=1),
        pydantic_v2.conint(gt=10, lt=100),
        pydantic_v2.confloat(ge=10, le=100),
        pydantic_v2.condecimal(gt=Decimal("10")),
        pydantic_v2.condate(gt=datetime.date.today()),
    ],
)
def test_is_pydantic_constrained_field(constrained: Any) -> None:
    PydanticSchemaPlugin.is_constrained_field(FieldDefinition.from_annotation(constrained))


def test_v2_constrained_secrets() -> None:
    # https://github.com/litestar-org/litestar/issues/3148
    class Model(pydantic_v2.BaseModel):
        string: pydantic_v2.SecretStr = pydantic_v2.Field(min_length=1)
        bytes_: pydantic_v2.SecretBytes = pydantic_v2.Field(min_length=1)

    schema = PydanticSchemaPlugin.for_pydantic_model(
        FieldDefinition.from_annotation(Model), schema_creator=SchemaCreator(plugins=[PydanticSchemaPlugin()])
    )
    assert schema.properties
    assert schema.properties["string"] == Schema(min_length=1, type=OpenAPIType.STRING)
    assert schema.properties["bytes_"] == Schema(min_length=1, type=OpenAPIType.STRING)


class V1ModelWithPrivateFields(pydantic_v1.BaseModel):
    class Config:
        underscore_fields_are_private = True

    _field: str = pydantic_v1.PrivateAttr()
    # include an invalid annotation here to ensure we never touch those fields
    _underscore_field: "foo"  # type: ignore[name-defined]  # noqa: F821


class V1GenericModelWithPrivateFields(pydantic_v1.generics.GenericModel, Generic[T]):  # pyright: ignore
    class Config:
        underscore_fields_are_private = True

    _field: str = pydantic_v1.PrivateAttr()
    # include an invalid annotation here to ensure we never touch those fields
    _underscore_field: "foo"  # type: ignore[name-defined]  # noqa: F821


class V2ModelWithPrivateFields(pydantic_v2.BaseModel):
    _field: str = pydantic_v2.PrivateAttr()
    # include an invalid annotation here to ensure we never touch those fields
    _underscore_field: "foo"  # type: ignore[name-defined] # noqa: F821


class V2GenericModelWithPrivateFields(pydantic_v2.BaseModel, Generic[T]):
    _field: str = pydantic_v2.PrivateAttr()
    # include an invalid annotation here to ensure we never touch those fields
    _underscore_field: "foo"  # type: ignore[name-defined] # noqa: F821


@pytest.mark.parametrize(
    "model_class",
    [
        V1ModelWithPrivateFields,
        V1GenericModelWithPrivateFields,
        V2ModelWithPrivateFields,
        V2GenericModelWithPrivateFields,
    ],
)
def test_exclude_private_fields(model_class: Type[Union[pydantic_v1.BaseModel, pydantic_v2.BaseModel]]) -> None:
    # https://github.com/litestar-org/litestar/issues/3150
    schema = PydanticSchemaPlugin.for_pydantic_model(
        FieldDefinition.from_annotation(model_class), schema_creator=SchemaCreator(plugins=[PydanticSchemaPlugin()])
    )
    assert not schema.properties



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Sequence

import msgspec
import yaml

from litestar.constants import OPENAPI_JSON_HANDLER_NAME
from litestar.enums import MediaType, OpenAPIMediaType
from litestar.handlers import get
from litestar.serialization import encode_json, get_serializer

if TYPE_CHECKING:
    from litestar.connection import Request
    from litestar.router import Router

__all__ = (
    "OpenAPIRenderPlugin",
    "RapidocRenderPlugin",
    "RedocRenderPlugin",
    "ScalarRenderPlugin",
    "StoplightRenderPlugin",
    "SwaggerRenderPlugin",
    "YamlRenderPlugin",
)

_favicon_url = "https://cdn.jsdelivr.net/gh/litestar-org/branding@main/assets/Branding%20-%20PNG%20-%20Transparent/Badge%20-%20Blue%20and%20Yellow.png"
_default_favicon = f"<link rel='icon' type='image/png' href='{_favicon_url}'>"
_default_style = "<style>body { margin: 0; padding: 0 }</style>"


class OpenAPIRenderPlugin(ABC):
    """Base class for OpenAPI UI render plugins."""

    paths: list[str]

    def __init__(
        self,
        *,
        path: str | Sequence[str],
        media_type: MediaType | OpenAPIMediaType = MediaType.HTML,
        favicon: str = _default_favicon,
        style: str = _default_style,
    ) -> None:
        """Initialize the OpenAPI UI render plugin.

        Args:
            path: Path to serve the OpenAPI UI at.
            media_type: Media type for the handler.
            favicon: Html <link> tag for the favicon.
            style: Base styling of the html body.
        """
        self.paths = [path] if isinstance(path, str) else list(path)
        self.media_type = media_type
        self.favicon = favicon
        self.style = style

    @staticmethod
    def render_json(request: Request, openapi_schema: dict[str, Any]) -> bytes:
        """Render the OpenAPI schema as JSON.

        Args:
            request: The request that triggered the render.
            openapi_schema: The OpenAPI schema as a dictionary.

        Returns:
            The rendered JSON.
        """
        return encode_json(openapi_schema, serializer=get_serializer(request.route_handler.resolve_type_encoders()))

    @abstractmethod
    def render(self, request: Request, openapi_schema: dict[str, Any]) -> bytes:
        """Render the OpenAPI UI.

        Args:
            request: The request that triggered the render.
            openapi_schema: The OpenAPI schema as a dictionary.

        Returns:
            The rendered HTML.
        """
        raise NotImplementedError

    @staticmethod
    def get_openapi_json_route(request: Request) -> str:
        """Get the route for the OpenAPI JSON schema.

        Returns:
            The route for the OpenAPI JSON schema.
        """
        return request.app.route_reverse(OPENAPI_JSON_HANDLER_NAME)

    def receive_router(self, router: Router) -> None:
        """Receive the router that serves the OpenAPI UI.

        Can be used by plugins to additionally configure the router, e.g. to add
        additional routes.

        Args:
            router: The router that serves the OpenAPI UI.
        """
        return

    def has_path(self, path: str) -> bool:
        """Check if the plugin has a path.

        Args:
            path: The path to check.

        Returns:
            True if the plugin has the path, False otherwise.
        """
        return path in self.paths


class JsonRenderPlugin(OpenAPIRenderPlugin):
    """Render the OpenAPI schema as JSON."""

    def __init__(
        self,
        *,
        path: str | Sequence[str] = "/openapi.json",
        media_type: MediaType | OpenAPIMediaType = OpenAPIMediaType.OPENAPI_JSON,
        **kwargs: Any,
    ) -> None:
        """Initialize the OpenAPI UI render plugin.

        Args:
            path: Path to serve the OpenAPI UI at.
            media_type: Media type for the handler.
            **kwargs: Additional arguments to pass to the base class.

        """
        super().__init__(path=path, media_type=media_type, **kwargs)

    def render(self, request: Request, openapi_schema: dict[str, Any]) -> bytes:
        """Render an OpenAPI schema as JSON.

        Args:
            request: The request.
            openapi_schema: The OpenAPI schema as a dictionary.

        Returns:
            The rendered OpenAPI schema as JSON.
        """
        return self.render_json(request, openapi_schema)


class YamlRenderPlugin(OpenAPIRenderPlugin):
    """Render an OpenAPI schema as YAML."""

    def __init__(
        self,
        *,
        path: str | Sequence[str] = ("/openapi.yaml", "/openapi.yml"),
        media_type: MediaType | OpenAPIMediaType = OpenAPIMediaType.OPENAPI_YAML,
        **kwargs: Any,
    ) -> None:
        """Initialize the OpenAPI UI render plugin.

        Args:
            path: Path to serve the OpenAPI UI at.
            media_type: Media type for the handler.
            **kwargs: Additional arguments to pass to the base class.
        """
        super().__init__(path=path, media_type=media_type, **kwargs)

    def render(self, request: Request, openapi_schema: dict[str, Any]) -> bytes:
        """Render an OpenAPI schema as YAML.

        Args:
            request: The request.
            openapi_schema: The OpenAPI schema as a dictionary.

        Returns:
            The rendered OpenAPI schema as YAML.
        """
        # using msgspec.to_builtins() ensures that any examples generated by polyfactory that have the
        # UNSET value (possible if the examples are being generated for a partial DTO model which makes
        # every type a union with UNSET) are stripped out.
        openapi_schema = msgspec.to_builtins(
            openapi_schema, enc_hook=get_serializer(request.route_handler.resolve_type_encoders())
        )
        return yaml.dump(openapi_schema, default_flow_style=False).encode("utf-8")


class RapidocRenderPlugin(OpenAPIRenderPlugin):
    """Render an OpenAPI schema using Rapidoc."""

    def __init__(
        self,
        *,
        version: str = "9.3.4",
        js_url: str | None = None,
        path: str | Sequence[str] = "/rapidoc",
        **kwargs: Any,
    ) -> None:
        """Initialize the OpenAPI UI render plugin.

        Args:
            version: Rapidoc version to download from the CDN. If js_url is provided, this is ignored.
            js_url: Download url for the RapiDoc JS bundle. If not provided, the version will be used to construct the
                url.
            path: Path to serve the OpenAPI UI at.
            **kwargs: Additional arguments to pass to the base class.
        """
        self.js_url = js_url or f"https://unpkg.com/rapidoc@{version}/dist/rapidoc-min.js"
        super().__init__(path=path, **kwargs)

    def render(self, request: Request, openapi_schema: dict[str, Any]) -> bytes:
        """Render an HTML page for Rapidoc.

        .. note:: Override this method to customize the template.

        Args:
            request: The request.
            openapi_schema: The OpenAPI schema as a dictionary.

        Returns:
            A rendered html string.
        """

        head = f"""
          <head>
            <title>{openapi_schema["info"]["title"]}</title>
            {self.favicon}
            <meta charset="utf-8"/>
            <meta name="viewport" content="width=device-width, initial-scale=1">
            <script src="{self.js_url}" crossorigin></script>
            {self.style}
          </head>
        """

        body = f"""
          <body>
            <rapi-doc spec-url="{self.get_openapi_json_route(request)}" />
          </body>
        """

        return f"""
        <!DOCTYPE html>
            <html>
                {head}
                {body}
            </html>
        """.encode()


class RedocRenderPlugin(OpenAPIRenderPlugin):
    """Render an OpenAPI schema using Redoc."""

    def __init__(
        self,
        *,
        version: str = "next",
        js_url: str | None = None,
        google_fonts: bool = True,
        path: str | Sequence[str] = "/redoc",
        **kwargs: Any,
    ) -> None:
        """Initialize the OpenAPI UI render plugin.

        Args:
            version: Redoc version to download from the CDN. If js_url is provided, this is ignored.
            js_url: Download url for the Redoc JS bundle. If not provided, the version will be used to construct the url.
            google_fonts: Download google fonts via CDN. Should be set to False when not using a CDN.
            path: Path to serve the OpenAPI UI at.
            **kwargs: Additional arguments to pass to the base class.
        """
        self.js_url = js_url or f"https://cdn.jsdelivr.net/npm/redoc@{version}/bundles/redoc.standalone.js"
        self.google_fonts = google_fonts
        super().__init__(path=path, **kwargs)

    def render(self, request: Request, openapi_schema: dict[str, Any]) -> bytes:
        """Render an HTML page for Redoc.

        .. note:: override this method to customize the template.

        Args:
            request: The request.
            openapi_schema: The OpenAPI schema as a dictionary.

        Returns:
            A rendered html string.
        """

        head = f"""
          <head>
            <title>{openapi_schema["info"]["title"]}</title>
            {self.favicon}
            <meta charset="utf-8"/>
            <meta name="viewport" content="width=device-width, initial-scale=1">
            """

        if self.google_fonts:
            head += """
            <link href="https://fonts.googleapis.com/css?family=Montserrat:300,400,700|Roboto:300,400,700" rel="stylesheet">
            """

        head += f"""
            <script src="{self.js_url}" crossorigin></script>
            {self.style}
          </head>
        """

        body = b"".join(
            [
                b"<body><div id='redoc-container'/><script type='text/javascript'>Redoc.init(",
                self.render_json(request, openapi_schema),
                b",undefined,document.getElementById('redoc-container'))</script></body>",
            ]
        )

        return b"".join(
            [
                b"<!DOCTYPE html><html>",
                head.encode(),
                body,
                b"</html>",
            ]
        )


class ScalarRenderPlugin(OpenAPIRenderPlugin):
    """Plugin to render an OpenAPI schema using Scalar.

    .. versionadded:: 2.8.0
    """

    _default_css_url = "https://cdn.jsdelivr.net/gh/litestar-org/branding@main/assets/openapi/scalar.css"

    def __init__(
        self,
        *,
        version: str = "1.19.5",
        js_url: str | None = None,
        css_url: str | None = None,
        path: str | Sequence[str] = "/scalar",
        **kwargs: Any,
    ) -> None:
        """Initialize the Scalar OpenAPI UI render plugin.

        Args:
            version: Scalar version to download from the CDN.
                If js_url is provided, this is ignored.
            js_url: Download url for the Scalar JS bundle.
                If not provided, the version will be used to construct the url.
            css_url: Download url for the Scalar CSS bundle.
                If not provided, the Litestar-provided CSS will be used.
            path: Path to serve the OpenAPI UI at.
            **kwargs: Additional arguments to pass to the base class.
        """
        self.js_url = js_url or f"https://cdn.jsdelivr.net/npm/@scalar/api-reference@{version}"
        self.css_url = css_url or self._default_css_url
        super().__init__(path=path, **kwargs)

    def render(self, request: Request, openapi_schema: dict[str, Any]) -> bytes:
        """Render an HTMl page for Scalar.

        .. note:: Override this method to customize the template.

        Args:
            request: The request.
            openapi_schema: The OpenAPI schema as a dictionary.

        Returns:
            A rendered html string.
        """
        head = f"""
                  <head>
                    <title>{openapi_schema["info"]["title"]}</title>
                    {self.style}
                    <meta charset="utf-8"/>
                    <meta name="viewport" content="width=device-width, initial-scale=1">
                    {self.favicon}
                    <link rel="stylesheet" type="text/css" href="{self.css_url}">
                  </head>
                """

        body = f"""
                <noscript>
                    Scalar requires Javascript to function. Please enable it to browse the documentation.
                </noscript>
                <script
                  id="api-reference"
                  data-url="{self.get_openapi_json_route(request)}">
                </script>
                <script src="{self.js_url}" crossorigin></script>
                """

        return f"""
                <!DOCTYPE html>
                    <html>
                        {head}
                        {body}
                    </html>
                """.encode()


class StoplightRenderPlugin(OpenAPIRenderPlugin):
    """Render an OpenAPI schema using StopLight Elements."""

    def __init__(
        self,
        *,
        version: str = "7.7.18",
        js_url: str | None = None,
        css_url: str | None = None,
        path: str | Sequence[str] = "/elements",
        **kwargs: Any,
    ) -> None:
        """Initialize the OpenAPI UI render plugin.

        Args:
            version: StopLight Elements version to download from the CDN. If js_url is provided, this is ignored.
            js_url: Download url for the StopLight Elements JS bundle. If not provided, the version will be used to
                construct the url.
            css_url: Download url for the StopLight Elements CSS bundle. If not provided, the version will be used to
                construct the url.
            path: Path to serve the OpenAPI UI at.
            **kwargs: Additional arguments to pass to the base class.
        """
        self.js_url = js_url or f"https://unpkg.com/@stoplight/elements@{version}/web-components.min.js"
        self.css_url = css_url or f"https://unpkg.com/@stoplight/elements@{version}/styles.min.css"
        super().__init__(path=path, **kwargs)

    def render(self, request: Request, openapi_schema: dict[str, Any]) -> bytes:
        """Render an HTML page for StopLight Elements.

        .. note:: Override this method to customize the template.

        Args:
            request: The request.
            openapi_schema: The OpenAPI schema as a dictionary.

        Returns:
            A rendered html string.
        """
        head = f"""
          <head>
            <title>{openapi_schema["info"]["title"]}</title>
            {self.favicon}
            <meta charset="utf-8"/>
            <meta name="viewport" content="width=device-width, initial-scale=1, shrink-to-fit=no">
            <link rel="stylesheet" href="{self.css_url}">
            <script src="{self.js_url}" crossorigin></script>
            {self.style}
          </head>
        """

        body = f"""
          <body>
            <elements-api
                apiDescriptionUrl="{self.get_openapi_json_route(request)}"
                router="hash"
                layout="sidebar"
            />
          </body>
        """

        return f"""
        <!DOCTYPE html>
            <html>
                {head}
                {body}
            </html>
        """.encode()


class SwaggerRenderPlugin(OpenAPIRenderPlugin):
    """Render an OpenAPI schema using Swagger-UI."""

    def __init__(
        self,
        version: str = "5.1.3",
        js_url: str | None = None,
        css_url: str | None = None,
        standalone_preset_js_url: str | None = None,
        init_oauth: dict[str, Any] | bytes | None = None,
        path: str | Sequence[str] = "/swagger",
        **kwargs: Any,
    ) -> None:
        """Initialize the OpenAPI UI render plugin.

        Args:
            version: SwaggerUI version to download from the CDN. If js_url is provided, this is ignored.
            js_url: Download url for the Swagger UI JS bundle. If not provided, the version will be used to construct
                the url.
            css_url: Download url for the Swagger UI CSS bundle. If not provided, the version will be used to construct
                the url.
            standalone_preset_js_url: Download url for the Swagger Standalone Preset JS bundle. If not provided, the
                version will be used to construct the url.
            init_oauth: JSON to initialize Swagger UI OAuth2 by calling the ``initOAuth`` method.
                Refer to the following URL for details:
                `Swagger-UI <https://swagger.io/docs/open-source-tools/swagger-ui/usage/oauth2/>`_.
            path: Path to serve the OpenAPI UI at.
            **kwargs: Additional arguments to pass to the base class.
        """
        self.js_url = js_url or f"https://cdn.jsdelivr.net/npm/swagger-ui-dist@{version}/swagger-ui-bundle.js"
        self.css_url = css_url or f"https://cdn.jsdelivr.net/npm/swagger-ui-dist@{version}/swagger-ui.css"
        self.standalone_preset_js_url = (
            standalone_preset_js_url
            or f"https://cdn.jsdelivr.net/npm/swagger-ui-dist@{version}/swagger-ui-standalone-preset.js"
        )
        self.init_oauth = init_oauth or {}
        super().__init__(path=path, **kwargs)

    def render(self, request: Request, openapi_schema: dict[str, Any]) -> bytes:
        """Render an HTML page for Swagger-UI.

        Notes:
            - override this method to customize the template.

        Args:
            request: The request.
            openapi_schema: The OpenAPI schema as a dictionary.

        Returns:
            A rendered html string.
        """

        head = f"""
          <head>
            <title>{openapi_schema["info"]["title"]}</title>
            {self.favicon}
            <meta charset="utf-8"/>
            <meta name="viewport" content="width=device-width, initial-scale=1">
            <link href="{self.css_url}" rel="stylesheet">
            <script src="{self.js_url}" crossorigin></script>
            <script src="{self.standalone_preset_js_url}" crossorigin></script>
            {self.style}
          </head>
        """

        body = b"".join(
            [
                b"""
            <body>
              <div id='swagger-container'/>
                <script type='text/javascript'>
                const ui = SwaggerUIBundle({
                  spec: """,
                self.render_json(request, openapi_schema),
                b""",
                  dom_id: '#swagger-container',
                  deepLinking: true,
                  showExtensions: true,
                  showCommonExtensions: true,
                  presets: [
                      SwaggerUIBundle.presets.apis,
                      SwaggerUIBundle.SwaggerUIStandalonePreset
                  ],
                })
            ui.initOAuth(""",
                encode_json(self.init_oauth),
                b""")
            </script>
          </body>
        """,
            ]
        )

        return b"".join([b"<!DOCTYPE html><html>", head.encode(), body, b"</html>"])

    def receive_router(self, router: Router) -> None:
        """Receive the router that serves the OpenAPI UI.

        Adds a route to serve the OAuth2 redirect page.

        Args:
            router: The router that serves the OpenAPI UI.
        """
        router.register(
            get("/oauth2-redirect.html", media_type=MediaType.HTML, sync_to_thread=False)(self.render_oauth2_redirect),
        )

    @staticmethod
    def render_oauth2_redirect() -> bytes:
        """Render an HTML oauth2-redirect.html page for Swagger-UI.

        .. note:: Override this method to customize the template.

        Returns:
            A rendered html string.
        """
        return rb"""<!doctype html>
        <html lang="en-US">
        <head>
            <title>Swagger UI: OAuth2 Redirect</title>
        </head>
        <body>
        <script>
            'use strict';
            function run () {
                var oauth2 = window.opener.swaggerUIRedirectOauth2;
                var sentState = oauth2.state;
                var redirectUrl = oauth2.redirectUrl;
                var isValid, qp, arr;

                if (/code|token|error/.test(window.location.hash)) {
                    qp = window.location.hash.substring(1).replace('?', '&');
                } else {
                    qp = location.search.substring(1);
                }

                arr = qp.split("&");
                arr.forEach(function (v,i,_arr) { _arr[i] = '"' + v.replace('=', '":"') + '"';});
                qp = qp ? JSON.parse('{' + arr.join() + '}',
                        function (key, value) {
                            return key === "" ? value : decodeURIComponent(value);
                        }
                ) : {};

                isValid = qp.state === sentState;

                if ((
                oauth2.auth.schema.get("flow") === "accessCode" ||
                oauth2.auth.schema.get("flow") === "authorizationCode" ||
                oauth2.auth.schema.get("flow") === "authorization_code"
                ) && !oauth2.auth.code) {
                    if (!isValid) {
                        oauth2.errCb({
                            authId: oauth2.auth.name,
                            source: "auth",
                            level: "warning",
                            message: "Authorization may be unsafe, passed state was changed in server. The passed state wasn't returned from auth server."
                        });
                    }

                    if (qp.code) {
                        delete oauth2.state;
                        oauth2.auth.code = qp.code;
                        oauth2.callback({auth: oauth2.auth, redirectUrl: redirectUrl});
                    } else {
                        let oauthErrorMsg;
                        if (qp.error) {
                            oauthErrorMsg = "["+qp.error+"]: " +
                                (qp.error_description ? qp.error_description+ ". " : "no accessCode received from the server. ") +
                                (qp.error_uri ? "More info: "+qp.error_uri : "");
                        }

                        oauth2.errCb({
                            authId: oauth2.auth.name,
                            source: "auth",
                            level: "error",
                            message: oauthErrorMsg || "[Authorization failed]: no accessCode received from the server."
                        });
                    }
                } else {
                    oauth2.callback({auth: oauth2.auth, token: qp, isValid: isValid, redirectUrl: redirectUrl});
                }
                window.close();
            }

            if (document.readyState !== 'loading') {
                run();
            } else {
                document.addEventListener('DOMContentLoaded', function () {
                    run();
                });
            }
        </script>
        </body>
        </html>"""



from __future__ import annotations

from dataclasses import dataclass

from litestar.openapi.spec.base import BaseSchemaObject

__all__ = ("ServerVariable",)


@dataclass
class ServerVariable(BaseSchemaObject):
    """An object representing a Server Variable for server URL template substitution."""

    default: str
    """**REQUIRED**. The default value to use for substitution, which SHALL be sent if an alternate value is _not_
    supplied. Note this behavior is different than the
    `Schema Object's <https://spec.openapis.org/oas/v3.1.0#schemaObject>`_ treatment of default values, because in those
    cases parameter values are optional.  If the `enum <https://spec.openapis.org/oas/v3.1.0#serverVariableEnum>`_ is
    defined, the value MUST exist in the enum's values.
    """

    enum: list[str] | None = None
    """An enumeration of string values to be used if the substitution options are from a limited set.

    The array SHOULD NOT be empty.
    """

    description: str | None = None
    """An optional description for the server variable.

    `CommonMark syntax <https://spec.commonmark.org/>`_ MAY be used for rich text representation.
    """



from __future__ import annotations

from typing import TYPE_CHECKING, Dict

if TYPE_CHECKING:
    from litestar.openapi.spec import PathItem

Paths = Dict[str, "PathItem"]
"""Holds the relative paths to the individual endpoints and their operations. The path is appended to the URL from the.

`Server Object <https://spec.openapis.org/oas/v3.1.0#serverObject>`_ in order to construct the full URL.

The Paths MAY be empty, due to
`Access Control List (ACL) constraints <https://spec.openapis.org/oas/v3.1.0#securityFiltering>`_.

Patterned Fields

/{path}: PathItem

A relative path to an individual endpoint. The field name MUST begin with a forward slash (``/``). The path is
**appended** (no relative URL resolution) to the expanded URL from the
`Server Object <https://spec.openapis.org/oas/v3.1.0#serverObject>`_ 's ``url`` field in order to construct the full
URL. `Path templating <https://spec.openapis.org/oas/v3.1.0#pathTemplating>`_ is allowed. When matching URLs, concrete
(non-templated) paths would be matched before their templated counterparts. Templated paths with the same hierarchy but
different templated names MUST NOT exist as they are identical. In case of ambiguous matching, it's up to the tooling to
decide which one to use.
"""
